@@ -1,10 +1,11 @@
+"""Machine Learning / Time series Loss and Evaluation Metrics"""
+
 from __future__ import annotations
 
 import polars as pl
 
-from ._utils import pl_plugin
-from .type_alias import MultiAUCStrategy, str_to_expr
-
+from ._utils import pl_plugin, str_to_expr
+from .typing import MultiAUCStrategy
 
 __all__ = [
     "query_r2",
@@ -21,12 +22,19 @@ __all__ = [
     "query_msle",
     "query_mad",
     "query_roc_auc",
+    "query_tpr_fpr",
     "query_binary_metrics",
     "query_multi_roc_auc",
     "query_cat_cross_entropy",
     "query_confusion_matrix",
+    "query_fairness",
+    "query_p_pct_score",
+    "query_mcc",
 ]
 
+# Confusion matrix based metrics should all be covered. If there is no
+# specific function for it, we can find it in query_confusion_matrix.
+
 
 def query_mad(x: str | pl.Expr, use_mean: bool = True) -> pl.Expr:
     """
@@ -189,13 +197,22 @@ def query_log_loss(actual: str | pl.Expr, pred: str | pl.Expr, normalize: bool =
     normalize
         Whether to divide by N.
     """
-    a = str_to_expr(actual)
-    p = str_to_expr(pred)
-    out = a.dot(p.log()) + (1 - a).dot((1 - p).log())
-    if normalize:
-        return -(out / a.count())
-    return -out
+    a = str_to_expr(actual).cast(pl.Float64)
+    p = str_to_expr(pred).cast(pl.Float64)
+    first = pl_plugin(
+        args=[a, p],
+        symbol="pl_xlogy",
+        is_elementwise=True,
+    )
+    second = pl_plugin(
+        args=[pl.lit(1.0, dtype=pl.Float64) - a, pl.lit(1.0, dtype=pl.Float64) - p],
+        symbol="pl_xlogy",
+        is_elementwise=True,
+    )
 
+    if normalize:
+        return -(first + second).mean()
+    return -(first + second).sum()
 
 def query_mape(actual: str | pl.Expr, pred: str | pl.Expr, weighted: bool = False) -> pl.Expr:
     """
@@ -343,7 +360,27 @@ def query_roc_auc(
         symbol="pl_roc_auc",
         args=[str_to_expr(actual).cast(pl.UInt32), str_to_expr(pred)],
         returns_scalar=True,
-        pass_name_to_apply=True,
+    )
+
+
+def query_tpr_fpr(
+    actual: str | pl.Expr,
+    pred: str | pl.Expr,
+) -> pl.Expr:
+    """
+    Returns the TPR and FPR for all thresholds. This is useful when you want to study the thresholds
+    or when you want to plot roc auc curve.
+
+    Parameters
+    ----------
+    actual
+        An expression represeting the actual
+    pred
+        An expression represeting the column with predicted probability.
+    """
+    return pl_plugin(
+        symbol="pl_tpr_fpr",
+        args=[str_to_expr(actual).cast(pl.UInt32), str_to_expr(pred)],
     )
 
 
@@ -546,3 +583,74 @@ def query_cat_cross_entropy(
     if normalize:
         return -y_prob.log().sum() / a.count()
     return -y_prob.log().sum()
+
+
+def query_mcc(y_true: str | pl.Expr, y_pred: str | pl.Expr) -> pl.Expr:
+    """
+    Returns the Matthews correlation coefficient (phi coefficient). The inputs must be 0s and 1s
+    and castable to u32. If not, the result may not be correct. See query_confusion_matrix for querying
+    all the confusion metrics at the same time.
+
+    Parameters
+    ----------
+    y_true
+        The true labels. Must be 0s and 1s.
+    y_pred
+        The predicted labels. Must be 0s and 1s. E.g. This could be say (y_prob > 0.5).cast(pl.UInt32)
+
+    Reference
+    ---------
+    https://en.wikipedia.org/wiki/Phi_coefficient
+    """
+
+    y = str_to_expr(y_true)
+    x = str_to_expr(y_pred)
+    combined = (2 * y + x).cast(pl.UInt32)
+
+    return pl_plugin(
+        symbol="pl_mcc",
+        args=[combined],
+        returns_scalar=True,
+    )
+
+
+def query_fairness(pred: str | pl.Expr, sensitive_cond: pl.Expr) -> pl.Expr:
+    """
+    A simple fairness metric for regression output. Computes the absolute difference between
+    the average of the `pred` values on when the `sensitive_cond` is true vs the
+    avg of the values when `sensitive_cond` is false.
+
+    The lower this value is, the more fair is the model on the sensitive condition.
+
+    Parameters
+    ----------
+    pred
+        The predictions
+    sensitive_cond
+        A boolean expression representing the sensitive condition
+    """
+    p = str_to_expr(pred)
+    return (p.filter(sensitive_cond).mean() - p.filter(~sensitive_cond).mean()).abs()
+
+
+def query_p_pct_score(pred: str | pl.Expr, sensitive_cond: pl.Expr) -> pl.Expr:
+    """
+    Computes the 'p-percent score', which measures the fairness of a classification
+    model on a sensitive_cond. Let z = the sensitive_cond, then:
+
+    p-percent score = min(P(y = 1 | z = 1) / P(y = 1 | z = 0), P(y = 1 | z = 0) / P(y = 1 | z = 1))
+
+    Parameters
+    ----------
+    pred
+        The predictions. Must be 0s and 1s.
+    sensitive_cond
+        A boolean expression representing the sensitive condition
+    """
+    p = str_to_expr(pred)
+    p_y1_z1 = p.filter(
+        sensitive_cond
+    ).mean()  # since p is 0s and 1s, this is equal to P(pred = 1 | sensitive_cond)
+    p_y1_z0 = p.filter(~sensitive_cond).mean()
+    ratio = p_y1_z1 / p_y1_z0
+    return pl.min_horizontal(ratio, 1 / ratio)
