@@ -1,32 +1,33 @@
+"""Miscallaneous Numerical Functions and Transforms."""
+
 from __future__ import annotations
 import math
 import polars as pl
-from typing import Union, List, Iterable
-from .type_alias import (
+from typing import List, Iterable
+from .typing import (
     DetrendMethod,
     ConvMode,
     ConvMethod,
-    str_to_expr,
 )
-from ._utils import pl_plugin
+from ._utils import pl_plugin, str_to_expr
 
 __all__ = [
-    "query_singular_values",
-    "query_principal_components",
+    "singular_values",
+    "principal_components",
+    "pca",
     "softmax",
-    "query_gcd",
-    "query_lcm",
+    "gcd",
+    "lcm",
     "haversine",
-    "query_pca",
-    "query_jaccard_row",
-    "query_jaccard_col",
-    "query_psi",
-    "query_psi_w_breakpoints",
-    "query_psi_discrete",
-    "query_woe",
-    "query_woe_discrete",
-    "query_iv",
-    "query_iv_discrete",
+    "jaccard_row",
+    "jaccard_col",
+    "psi",
+    "psi_w_breakpoints",
+    "psi_discrete",
+    "woe",
+    "woe_discrete",
+    "info_value",
+    "info_value_discrete",
     "integrate_trapz",
     "convolve",
     "list_amax",
@@ -39,8 +40,132 @@ __all__ = [
     "detrend",
     "rfft",
     "fract",
+    "center",
+    "z_normalize",
+    "isotonic_regression",
+    "is_increasing",
+    "is_decreasing",
+    "next_up",
+    "next_down",
+    "digamma",
+    "xlogy",
+    "l_inf_horizontal",
+    "l1_horizontal",
+    "l2_sq_horizontal"
+    # "mutual_info_disc",
 ]
 
+def l_inf_horizontal(*v: str | pl.Expr, normalize: bool = False) -> pl.Expr:
+    """
+    Horizontally computes L inf norm. Shorthand for pl.max_horizontal(pl.col(x).abs() for x in exprs).
+
+    Parameters
+    ----------
+    *v
+        Expressions to compute horizontal L infinity.
+    normalize
+        Whether to divide by the dimension
+    """
+    if normalize:
+        exprs = list(v)
+        return pl.max_horizontal(str_to_expr(x).abs() for x in exprs) / len(exprs)
+    else:
+        return pl.max_horizontal(str_to_expr(x).abs() for x in v)
+
+
+def l2_sq_horizontal(*v: str | pl.Expr, normalize: bool = False) -> pl.Expr:
+    """
+    Horizontally computes L2 norm squared. Shorthand for pl.sum_horizontal(pl.col(x).pow(2) for x in exprs).
+
+    Parameters
+    ----------
+    *v
+        Expressions to compute horizontal L2.
+    normalize
+        Whether to divide by the dimension
+    """
+    if normalize:
+        exprs = list(v)
+        return pl.sum_horizontal(str_to_expr(x).pow(2) for x in exprs) / len(exprs)
+    else:
+        return pl.sum_horizontal(str_to_expr(x).pow(2) for x in v)
+
+
+def l1_horizontal(*v: str | pl.Expr, normalize: bool = False) -> pl.Expr:
+    """
+    Horizontally computes L1 norm. Shorthand for pl.sum_horizontal(pl.col(x).abs() for x in exprs).
+
+    Parameters
+    ----------
+    *v
+        Expressions to compute horizontal L1.
+    normalize
+        Whether to divide by the dimension
+    """
+    if normalize:
+        exprs = list(v)
+        return pl.sum_horizontal(str_to_expr(x).abs() for x in exprs) / len(exprs)
+    else:
+        return pl.sum_horizontal(str_to_expr(x).abs() for x in v)
+
+def is_increasing(x: str | pl.Expr, strict: bool = False) -> pl.Expr:
+    """
+    Checks whether the column is monotonically increasing.
+
+    Parameters
+    ----------
+    x
+        A numerical column
+    strict
+        Whether the check should be strict
+    """
+    if strict:
+        return (str_to_expr(x).diff() > 0.0).all()
+    else:
+        return (str_to_expr(x).diff() >= 0.0).all()
+
+
+def is_decreasing(x: str | pl.Expr, strict: bool = False) -> pl.Expr:
+    """
+    Checks whether the column is monotonically decreasing.
+
+    Parameters
+    ----------
+    x
+        A numerical column
+    strict
+        Whether the check should be strict
+    """
+    xx = str_to_expr(x)
+    if strict:
+        return (xx.diff() < 0.0).all()
+    else:
+        return (xx.diff() <= 0.0).all()
+
+
+def center(x: str | pl.Expr) -> pl.Expr:
+    """
+    Centers the column.
+
+    This is only a short cut for a standard feature transform, and is not recommended
+    to be used in settings where the means need to be persisted.
+    """
+    xx = str_to_expr(x)
+    return xx - xx.mean()
+
+
+def z_normalize(x: str | pl.Expr) -> pl.Expr:
+    """
+    Z-normalizes the column.
+
+    This is only a short cut for a standard feature transform, and is not recommended
+    to be used in settings where the means/stds need to be persisted.
+    """
+    xx = str_to_expr(x)
+    mean = xx.mean()
+    std = xx.std()
+    return (xx - mean) / std
+
 
 def softmax(x: str | pl.Expr) -> pl.Expr:
     """
@@ -56,7 +181,7 @@ def softmax(x: str | pl.Expr) -> pl.Expr:
     return xx.exp() / (xx.exp().sum())
 
 
-def query_gcd(x: str | pl.Expr, y: int | str | pl.Expr) -> pl.Expr:
+def gcd(x: str | pl.Expr, y: int | str | pl.Expr) -> pl.Expr:
     """
     Computes GCD of two integer columns. This will try to cast everything to int32.
 
@@ -79,7 +204,7 @@ def query_gcd(x: str | pl.Expr, y: int | str | pl.Expr) -> pl.Expr:
     )
 
 
-def query_lcm(x: str | pl.Expr, y: Union[int, str, pl.Expr]) -> pl.Expr:
+def lcm(x: str | pl.Expr, y: int | str | pl.Expr) -> pl.Expr:
     """
     Computes LCM of two integer columns. This will try to cast everything to int32.
 
@@ -105,8 +230,8 @@ def query_lcm(x: str | pl.Expr, y: Union[int, str, pl.Expr]) -> pl.Expr:
 def haversine(
     x_lat: str | pl.Expr,
     x_long: str | pl.Expr,
-    y_lat: Union[float, str, pl.Expr],
-    y_long: Union[float, str, pl.Expr],
+    y_lat: float | str | pl.Expr,
+    y_long: float | str | pl.Expr,
 ) -> pl.Expr:
     """
     Computes haversine distance using the naive method. The output unit is km.
@@ -134,7 +259,7 @@ def haversine(
     )
 
 
-def query_singular_values(
+def singular_values(
     *features: str | pl.Expr,
     center: bool = True,
     as_explained_var: bool = False,
@@ -173,7 +298,7 @@ def query_singular_values(
     return out
 
 
-def query_pca(
+def pca(
     *features: str | pl.Expr,
     center: bool = True,
 ) -> pl.Expr:
@@ -194,16 +319,19 @@ def query_pca(
     else:
         actual_inputs = feats
 
-    return pl_plugin(symbol="pl_pca", args=actual_inputs, changes_length=True)
+    return pl_plugin(
+        symbol="pl_pca", args=actual_inputs, changes_length=True, pass_name_to_apply=True
+    )
 
 
-def query_principal_components(
+def principal_components(
     *features: str | pl.Expr,
     k: int = 2,
     center: bool = True,
 ) -> pl.Expr:
     """
-    Transforms the features to get the first k principal components.
+    Transforms the features to get the first k principal components. This returns NaN if the number
+    of rows is less than `k`.
 
     Paramters
     ---------
@@ -217,16 +345,16 @@ def query_principal_components(
     if k > len(feats) or k <= 0:
         raise ValueError("Input `k` should be between 1 and the number of features inclusive.")
 
-    actual_inputs = [pl.lit(k, dtype=pl.UInt32).alias("principal_components")]
+    actual_inputs = [pl.lit(k, dtype=pl.UInt32)]
     if center:
         actual_inputs.extend(f - f.mean() for f in feats)
     else:
         actual_inputs.extend(feats)
 
-    return pl_plugin(symbol="pl_principal_components", args=actual_inputs)
+    return pl_plugin(symbol="pl_principal_components", args=actual_inputs, pass_name_to_apply=True)
 
 
-def query_jaccard_row(first: str | pl.Expr, second: str | pl.Expr) -> pl.Expr:
+def jaccard_row(first: str | pl.Expr, second: str | pl.Expr) -> pl.Expr:
     """
     Computes jaccard similarity pairwise between this and the other column. The type of
     each column must be list and the lists must have the same inner type. The inner type
@@ -246,9 +374,7 @@ def query_jaccard_row(first: str | pl.Expr, second: str | pl.Expr) -> pl.Expr:
     )
 
 
-def query_jaccard_col(
-    first: str | pl.Expr, second: str | pl.Expr, count_null: bool = False
-) -> pl.Expr:
+def jaccard_col(first: str | pl.Expr, second: str | pl.Expr, count_null: bool = False) -> pl.Expr:
     """
     Computes jaccard similarity column-wise. This will hash entire columns and compares the two
     hashsets. Note: only integer/str columns can be compared.
@@ -269,7 +395,7 @@ def query_jaccard_col(
     )
 
 
-def query_psi(
+def psi(
     new: str | pl.expr | Iterable[float],
     baseline: str | pl.expr | Iterable[float],
     n_bins: int = 10,
@@ -312,17 +438,17 @@ def query_psi(
 
     if isinstance(new, (str, pl.Expr)):
         new_ = str_to_expr(new)
-        valid_new: Union[pl.Series, pl.Expr] = new_.filter(new_.is_finite()).cast(pl.Float64)
+        valid_new: pl.Series | pl.Expr = new_.filter(new_.is_finite()).cast(pl.Float64)
     else:
         temp = pl.Series(values=new, dtype=pl.Float64)
-        valid_new: Union[pl.Series, pl.Expr] = temp.filter(temp.is_finite())
+        valid_new: pl.Series | pl.Expr = temp.filter(temp.is_finite())
 
     if isinstance(baseline, (str, pl.Expr)):
         base = str_to_expr(baseline)
-        valid_ref: Union[pl.Series, pl.Expr] = base.filter(base.is_finite()).cast(pl.Float64)
+        valid_ref: pl.Series | pl.Expr = base.filter(base.is_finite()).cast(pl.Float64)
     else:
         temp = pl.Series(values=baseline, dtype=pl.Float64)
-        valid_ref: Union[pl.Series, pl.Expr] = temp.filter(temp.is_finite())
+        valid_ref: pl.Series | pl.Expr = temp.filter(temp.is_finite())
 
     vc = (
         valid_ref.qcut(n_bins, left_closed=False, allow_duplicates=True, include_breaks=True)
@@ -348,7 +474,7 @@ def query_psi(
     return psi_report.struct.field("psi_bin").sum()
 
 
-def query_psi_discrete(
+def psi_discrete(
     new: str | pl.expr | Iterable[float],
     baseline: str | pl.expr | Iterable[float],
     return_report: bool = False,
@@ -366,11 +492,11 @@ def query_psi_discrete(
 
     Parameters
     ----------
-    x
+    new
         The feature
     baseline
         An expression, or any iterable that can be turned into a Polars series. Usually this should
-        be x's historical values
+        be the historical values
     return_report
         Whether to return a PSI report or not.
 
@@ -381,24 +507,24 @@ def query_psi_discrete(
     if isinstance(new, (str, pl.Expr)):
         new_ = str_to_expr(new)
         temp = new_.value_counts().struct.rename_fields(["", "count"])
-        new_cnt: Union[pl.Series, pl.Expr] = temp.struct.field("count")
-        new_cat: Union[pl.Series, pl.Expr] = temp.struct.field("")
+        new_cnt: pl.Series | pl.Expr = temp.struct.field("count")
+        new_cat: pl.Series | pl.Expr = temp.struct.field("")
     else:
         temp = pl.Series(values=new)
         temp: pl.DataFrame = temp.value_counts()  # This is a df in this case
-        ref_cnt: Union[pl.Series, pl.Expr] = temp.drop_in_place("count")
-        ref_cat: Union[pl.Series, pl.Expr] = temp[temp.columns[0]]
+        ref_cnt: pl.Series | pl.Expr = temp.drop_in_place("count")
+        ref_cat: pl.Series | pl.Expr = temp[temp.columns[0]]
 
     if isinstance(baseline, (str, pl.Expr)):
         base = str_to_expr(baseline)
         temp = base.value_counts().struct.rename_fields(["", "count"])
-        ref_cnt: Union[pl.Series, pl.Expr] = temp.struct.field("count")
-        ref_cat: Union[pl.Series, pl.Expr] = temp.struct.field("")
+        ref_cnt: pl.Series | pl.Expr = temp.struct.field("count")
+        ref_cat: pl.Series | pl.Expr = temp.struct.field("")
     else:
         temp = pl.Series(values=baseline)
         temp: pl.DataFrame = temp.value_counts()  # This is a df in this case
-        ref_cnt: Union[pl.Series, pl.Expr] = temp.drop_in_place("count")
-        ref_cat: Union[pl.Series, pl.Expr] = temp[temp.columns[0]]
+        ref_cnt: pl.Series | pl.Expr = temp.drop_in_place("count")
+        ref_cat: pl.Series | pl.Expr = temp[temp.columns[0]]
 
     psi_report = pl_plugin(
         symbol="pl_psi_discrete_report",
@@ -411,7 +537,7 @@ def query_psi_discrete(
     return psi_report.struct.field("psi_bin").sum()
 
 
-def query_psi_w_breakpoints(
+def psi_w_breakpoints(
     new: str | pl.expr | Iterable[float],
     baseline: str | pl.expr | Iterable[float],
     breakpoints: List[float],
@@ -421,16 +547,17 @@ def query_psi_w_breakpoints(
 
     Parameters
     ----------
+    new
+        The data representing the new observed data. Any sequence of numerical values that
+        can be turned into a Polars'series, or an expression representing a column will work
     baseline
         The data representing the baseline data. Any sequence of numerical values that
         can be turned into a Polars'series, or an expression representing a column will work
-    actual
-        The data representing the actual, observed data. Any sequence of numerical values that
-        can be turned into a Polars'series, or an expression representing a column will work
     breakpoints
         The data that represents breakpoints. Input must be sorted, distinct, finite numeric values.
         This function will not cleanse the breakpoints for the user. E.g. [0.1, 0.5, 0.9] will create
-        four bins: (-inf. 0.1], (0.1, 0.5], (0.5, 0.9] and (0.9, inf).
+        four bins: (-inf. 0.1], (0.1, 0.5], (0.5, 0.9] and (0.9, inf). Please do not pass inf or NaN values
+        as breakpoints.
     """
     if isinstance(baseline, (str, pl.Expr)):
         x: pl.Expr = str_to_expr(baseline)
@@ -457,9 +584,7 @@ def query_psi_w_breakpoints(
     ).alias("psi_report")
 
 
-def query_woe(
-    x: str | pl.Expr, target: str | pl.expr | Iterable[float], n_bins: int = 10
-) -> pl.Expr:
+def woe(x: str | pl.Expr, target: str | pl.expr | Iterable[float], n_bins: int = 10) -> pl.Expr:
     """
     Compute the Weight of Evidence for x with respect to target. This assumes x
     is continuous. A value of 1 is added to all events/non-events
@@ -490,9 +615,9 @@ def query_woe(
     return pl_plugin(symbol="pl_woe_discrete", args=[brk, t], changes_length=True)
 
 
-def query_woe_discrete(
+def woe_discrete(
     x: str | pl.Expr,
-    target: Union[str | pl.Expr, Iterable[int]],
+    target: str | pl.Expr | Iterable[int],
 ) -> pl.Expr:
     """
     Compute the Weight of Evidence for x with respect to target. This assumes x
@@ -521,7 +646,7 @@ def query_woe_discrete(
     )
 
 
-def query_iv(
+def info_value(
     x: str | pl.Expr,
     target: str | pl.expr | Iterable[float],
     n_bins: int = 10,
@@ -561,7 +686,7 @@ def query_iv(
     return out.struct.field("iv").sum() if return_sum else out
 
 
-def query_iv_discrete(
+def info_value_discrete(
     x: str | pl.Expr, target: str | pl.Expr | Iterable[int], return_sum: bool = True
 ) -> pl.Expr:
     """
@@ -621,7 +746,7 @@ def integrate_trapz(y: str | pl.Expr, x: float | pl.Expr) -> pl.Expr:
 def convolve(
     x: str | pl.Expr,
     kernel: List[float] | "np.ndarray" | pl.Series | pl.Expr,  # noqa: F821
-    fill_value: Union[float, pl.Expr] = 0.0,
+    fill_value: float | pl.Expr = 0.0,
     method: ConvMethod = "direct",
     mode: ConvMode = "full",
     parallel: bool = False,
@@ -658,7 +783,7 @@ def convolve(
     https://en.wikipedia.org/wiki/Convolution
     """
     xx = str_to_expr(x).fill_null(fill_value).cast(pl.Float64).rechunk()  # One cont slice
-    f: Union[pl.Expr, pl.Series]
+    f: pl.Expr | pl.Series
     if isinstance(kernel, pl.Expr):
         f = kernel.filter(kernel.is_finite()).rechunk()  # One cont slice
     else:
@@ -763,6 +888,22 @@ def sinc(x: str | pl.Expr) -> pl.Expr:
     y = math.pi * pl.when(xx == 0).then(1e-20).otherwise(xx)
     return y.sin() / y
 
+def xlogy(x: str | pl.Expr, y: str | pl.Expr) -> pl.Expr:
+    """
+    Computes x * log(y) so that if x = 0, the product is 0.
+
+    Parameters
+    ----------
+    x
+        A numerical column
+    y
+        A numerical column
+    """
+    return pl_plugin(
+        args=[str_to_expr(x).cast(pl.Float64), str_to_expr(y).cast(pl.Float64)],
+        symbol="pl_xlogy",
+        is_elementwise=True,
+    )
 
 def detrend(x: str | pl.Expr, method: DetrendMethod = "linear") -> pl.Expr:
     """
@@ -847,3 +988,178 @@ def target_encode(
         kwargs={"min_samples_leaf": float(min_samples_leaf), "smoothing": smoothing},
         changes_length=True,
     )
+
+
+def isotonic_regression(
+    y: str | pl.Expr, weights: str | pl.Expr | None = None, increasing: bool = True
+) -> pl.Expr:
+    """
+    Performs isotonic regression on the data. This is the same as scipy.optimize.isotonic_regression.
+
+    Parameters
+    ----------
+    y
+        The response variable
+    weights
+        The weights for the response
+    increasing
+        If true, output will be monotonically inreasing. If false, it will be monotonically
+        decreasing.
+    """
+
+    yy = str_to_expr(y).cast(pl.Float64)
+    args = [yy]
+    has_weights = weights is not None
+    if has_weights:
+        args.append(str_to_expr(weights).cast(pl.Float64))
+
+    return pl_plugin(
+        symbol="pl_isotonic_regression",
+        args=args,
+        kwargs={
+            "has_weights": has_weights,
+            "increasing": increasing,
+        },
+    )
+
+
+def next_up(x: str | pl.Expr) -> pl.Expr:
+    """
+    For any float, return the least number greater than itself (within the precision).
+    Intergers will be treated as f32. E.g. The next value up for 0.1 is 0.10000000000000002
+    because of precision issues. This is useful when you need to make extremely small changes
+    to certain values and you don't want to add random noise.
+    """
+    return pl_plugin(
+        symbol="pl_next_up",
+        args=[str_to_expr(x)],
+        is_elementwise=True,
+    )
+
+
+def next_down(x: str | pl.Expr) -> pl.Expr:
+    """
+    For any float, return the greatest number smaller than itself (within the precision).
+    Intergers will be treated as f32. E.g. The next value down for 0.1 is 0.09999999999999999.
+    This is useful when you need to make extremely small changes to certain values and you don't
+    want to add random noise.
+    """
+    return pl_plugin(
+        symbol="pl_next_down",
+        args=[str_to_expr(x)],
+        is_elementwise=True,
+    )
+
+
+def digamma(x: str | pl.Expr) -> pl.Expr:
+    """
+    The diagamma function
+    """
+    return pl_plugin(
+        symbol="pl_diagamma",
+        args=[str_to_expr(x)],
+        is_elementwise=True,
+    )
+
+
+# Not sure what is wrong here
+
+# def mutual_info_disc(
+#     x: str | pl.Expr,
+#     target: str | pl.Expr,
+#     n_neighbors: int = 3,
+#     seed: int | None = None,
+# ) -> float:
+#     """
+#     Computes the mutual infomation between a continuous variable x and a discrete
+#     target varaible. Note: (1) This always assume `x` is continuous. (2) Unlike Scikit-learn,
+#     if a target category has <= `n_neighbors` records, then the result here may not make any sense.
+#     This is because (a) `n_neighbors` is typically a small number and if your target category has so
+#     few records, then maybe you need to rethink the target definition, and (b) if we assume target category
+#     always has > `n_neighbors` records, then we can speed up the algorithm by quite a bit.
+
+
+#     Reference
+#     ---------
+#     https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0087357
+#     """
+
+#     if n_neighbors <= 0:
+#         raise ValueError("Input `n_neighbors` must be > 0.")
+
+#     feat = str_to_expr(x).cast(pl.Float64)
+#     feat = feat / feat.std(ddof=0)
+#     t = str_to_expr(target)
+
+#     scale_factor = pl.max_horizontal(pl.lit(1.0, dtype=pl.Float64), feat.abs().mean())
+
+#     c = pl_plugin(
+#         symbol="pl_jitter",
+#         args=[feat, 1e-10 * scale_factor, pl.lit(seed, dtype=pl.UInt64)],
+#         is_elementwise=True,
+#     )
+
+#     kwargs = {
+#         "k": n_neighbors,
+#         "metric": "l1",
+#         "parallel": False,
+#         "skip_eval": False,
+#         "max_bound": 99999.0,
+#         "epsilon": 0.
+#     }
+
+#     # This is not really exposed to the user. It does `dist_from_kth_nb` and a `next_down` in one go.
+#     r = pl_plugin(
+#         symbol="pl_dist_from_kth_nb",
+#         args=[c],
+#         kwargs=kwargs,
+#     ).over(t)
+
+#     label_counts = c.len().over(t)
+
+#     # NB Cnt contains the point itself. This is what we want
+#     # This corresponds to m_i in the paper.
+#     nb_cnt = pl_plugin(
+#         symbol="pl_nb_cnt",
+#         args=[r, c],
+#         kwargs={
+#             "k": 0,
+#             "metric": "l1",  # Data is 1d, l2 metric is equal to l1, but l2 does more instructions. So use l1
+#             "parallel": False,
+#             "skip_eval": False,
+#             "skip_data": False,
+#         },
+#     )
+
+
+#     # psi in SciPy is the diagamma function
+#     psi_label_counts = pl_plugin(
+#         symbol="pl_diagamma",
+#         args=[label_counts],
+#         is_elementwise=True,
+#     )
+
+#     psi_nb_cnt = pl_plugin(
+#         symbol="pl_diagamma",
+#         args=[nb_cnt],
+#         is_elementwise=True,
+#     )
+
+#     psi_n_samples = pl_plugin(
+#         symbol="pl_diagamma",
+#         args=[pl.len()],
+#         is_elementwise=True,
+#     )
+
+#     psi_n_neighbors = pl_plugin(
+#         symbol="pl_diagamma",
+#         args=[pl.lit(n_neighbors)],
+#         is_elementwise=True,
+#     )
+
+#     return psi_n_samples + psi_n_neighbors - (psi_label_counts + psi_nb_cnt).mean()
+
+#     return pl.max_horizontal(
+#         pl.lit(0.0, dtype=pl.Float64),
+#         psi_n_samples + psi_n_neighbors - psi_label_counts.mean() - psi_nb_cnt.mean(),
+#     )
