@@ -1,3 +1,5 @@
+"""Tabular Machine Learning Pipelines with native Polars support."""
+
 from __future__ import annotations
 
 import polars as pl
@@ -8,20 +10,19 @@ from . import transforms as t
 from functools import partial
 from dataclasses import dataclass
 from polars.type_aliases import IntoExprColumn
-from typing import List, Union, Dict, Any
-from .type_alias import (
+from typing import List, Union, Dict, Any, Tuple
+from .typing import (
     TypeAlias,
     PolarsFrame,
     ExprTransform,
     FitTransformFunc,
     SimpleImputeMethod,
     SimpleScaleMethod,
-    StrOrExpr,
     QuantileMethod,
     EncoderDefaultStrategy,
 )
 
-from ._utils import _POLARS_LEGACY_SUPPORT, _IS_POLARS_V1
+from ._utils import _IS_POLARS_V1
 
 if sys.version_info >= (3, 11):
     from typing import Self
@@ -31,6 +32,14 @@ else:  # 3.10, 3.9, 3.8
 __all__ = ["Pipeline", "Blueprint", "FitStep"]
 
 
+@dataclass
+class SQLStep:  # FittedStep
+    sql: str
+
+    def __iter__(self):
+        return [self.sql].__iter__()
+
+
 @dataclass
 class SelectStep:  # FittedStep
     exprs: ExprTransform
@@ -58,28 +67,31 @@ class FilterStep:  # FittedStep
 @dataclass
 class FitStep:  # Not a FittedStep
     func: FitTransformFunc
-    cols: IntoExprColumn
+    cols: IntoExprColumn | None
     exclude: List[str]
 
     # Here we allow IntoExprColumn as input so that users can use selectors, or other polars expressions
     # to specify input columns, which adds flexibility.
     # We still need real column names so that the functions in transforms.py will work.
     def fit(self, df: PolarsFrame) -> ExprTransform:
-        if _IS_POLARS_V1:
-            real_cols: List[str] = [
-                x
-                for x in df.lazy().select(self.cols).collect_schema().names()
-                if x not in self.exclude
-            ]
+        if self.cols is None:
+            return self.func(df)
         else:
-            real_cols: List[str] = [
-                x for x in df.select(self.cols).columns if x not in self.exclude
-            ]
-        return self.func(df, real_cols)
+            if _IS_POLARS_V1:
+                real_cols: List[str] = [
+                    x
+                    for x in df.lazy().select(self.cols).collect_schema().names()
+                    if x not in self.exclude
+                ]
+            else:
+                real_cols: List[str] = [
+                    x for x in df.select(self.cols).columns if x not in self.exclude
+                ]
+            return self.func(df, real_cols)
 
 
-Step: TypeAlias = Union[FitStep, SelectStep, WithColumnsStep, FilterStep]
-FittedStep: TypeAlias = Union[SelectStep, WithColumnsStep, FilterStep]
+Step: TypeAlias = Union[FitStep, SelectStep, WithColumnsStep, FilterStep, SQLStep]
+FittedStep: TypeAlias = Union[SelectStep, WithColumnsStep, FilterStep, SQLStep]
 
 
 @dataclass
@@ -111,27 +123,32 @@ class StepRepr:
             raise ValueError(f"Keys missing or data type is not expected. Original error: \n{e}")
 
 
-def _to_json_dict(step: FittedStep) -> Dict:
+def _step_to_json(step: FittedStep) -> Dict:
     """
     Turns a fitted step into a JSON dict.
     """
-    if _IS_POLARS_V1:
-        try:
-            exprs = [e.meta.serialize(format="json") for e in step]
-        except Exception as e:
-            raise ValueError(f"The `FittedStep` is ill-defined. Original error: \n{e}")
+    if isinstance(step, SQLStep):
+        return {"SQLStep": step.sql}
     else:
-        try:
-            exprs = [e.meta.serialize() for e in step]
-        except Exception as e:
-            raise ValueError(f"The `FittedStep` is ill-defined. Original error: \n{e}")
+        if _IS_POLARS_V1:
+            try:
+                exprs = [e.meta.serialize(format="json") for e in step]
+            except Exception as e:
+                raise ValueError(f"The `FittedStep` is ill-defined. Original error: \n{e}")
+        else:
+            try:
+                exprs = [e.meta.serialize() for e in step]
+            except Exception as e:
+                raise ValueError(f"The `FittedStep` is ill-defined. Original error: \n{e}")
 
-    if isinstance(step, SelectStep):
-        return {"SelectStep": exprs}
-    elif isinstance(step, FilterStep):
-        return {"FilterStep": exprs}
-    else:
-        return {"WithColumnsStep": exprs}
+        if isinstance(step, SelectStep):
+            return {"SelectStep": exprs}
+        elif isinstance(step, FilterStep):
+            return {"FilterStep": exprs}
+        elif isinstance(step, WithColumnsStep):
+            return {"WithColumnsStep": exprs}
+        else:
+            raise ValueError("Unknown Pipeline Step Type.")
 
 
 @dataclass
@@ -156,7 +173,10 @@ class Pipeline:
         text: str = "Naive Query Steps: \n\n"
         for i, step in enumerate(self.transforms):
             text += f"Step {i+1}:\n"
-            text += ",\n".join(str(e) for e in step)  # SelectStep, WithColumnsStep are iterable
+            if isinstance(step, SQLStep):
+                text += f"Run SQL: {step.sql}\n"
+            else:
+                text += ",\n".join(str(e) for e in step)  # SelectStep, WithColumnsStep are iterable
             text += "\n\n"
 
         return text
@@ -179,6 +199,8 @@ class Pipeline:
                 plan = plan.select(step.exprs)
             elif isinstance(step, FilterStep):
                 plan = plan.filter(step.exprs)
+            elif isinstance(step, SQLStep):
+                plan = pl.SQLContext(df=plan, eager=False).execute(step.sql)
             else:
                 raise ValueError(f"Transform is not a valid FittedStep: {str(step)}")
 
@@ -188,15 +210,12 @@ class Pipeline:
         """
         Converts self to a dict, with all expressions turned into JSON strings.
         """
-        if _POLARS_LEGACY_SUPPORT:
-            raise ValueError("This functionality doesn't support polars < 0.20.16.")
-
         return {
             "name": str(self.name),
             "target": self.target,
             "feature_names_in_": list(self.feature_names_in_),
             "feature_names_out_": list(self.feature_names_out_),
-            "transforms": [_to_json_dict(step) for step in self.transforms],
+            "transforms": [_step_to_json(step) for step in self.transforms],
             "ensure_features_in": self.ensure_features_in,
             "ensure_features_out": self.ensure_features_out,
         }
@@ -213,9 +232,6 @@ class Pipeline:
         kwargs
             Keyword arguments to Python's default json
         """
-        if _POLARS_LEGACY_SUPPORT:
-            raise ValueError("This functionality doesn't support polars < 0.20.16.")
-
         # Maybe support other json package?
         if path is None:
             return json.dumps(self.to_dict(), **kwargs)
@@ -229,9 +245,6 @@ class Pipeline:
         """
         Recreates a pipeline from a dictionary created by the `to_dict` call.
         """
-        if _POLARS_LEGACY_SUPPORT:
-            raise ValueError("This functionality doesn't support polars < 0.20.16.")
-
         from io import StringIO
 
         try:
@@ -283,6 +296,9 @@ class Pipeline:
                     json_exprs = step.pop("FilterStep")
                     actual_exprs = [pl.Expr.deserialize(StringIO(e)) for e in json_exprs]
                     transform_steps.append(FilterStep(actual_exprs))
+            elif "SQLStep" in step:
+                sql = step.pop("SQLStep")
+                transform_steps.append(SQLStep(sql))
             else:
                 raise ValueError(f"Invalid step {step}")
 
@@ -300,16 +316,12 @@ class Pipeline:
         """
         Creates the Pipeline from the JSON string.
         """
-        if _POLARS_LEGACY_SUPPORT:
-            raise ValueError("This functionality doesn't support polars < 0.20.16.")
         return Pipeline.from_dict(json.loads(json_str))
 
     def from_json(path: str) -> Self:
         """
         Creates the Pipeline by loading a local JSON file at path
         """
-        if _POLARS_LEGACY_SUPPORT:
-            raise ValueError("This functionality doesn't support polars < 0.20.16.")
         with open(path, "r") as f:
             pipe_dict = json.load(f)
         return Pipeline.from_dict(pipe_dict)
@@ -332,7 +344,11 @@ class Pipeline:
         self.ensure_features_out = ensure_out
         return self
 
-    def transform(self, df: PolarsFrame, return_lazy: bool = False) -> PolarsFrame:
+    def transform(self, 
+        df: PolarsFrame, 
+        return_lazy: bool = False,
+        separate: bool = False,
+    ) -> PolarsFrame | Tuple[PolarsFrame, PolarsFrame]:
         """
         Transforms the df using the learned expressions.
 
@@ -343,6 +359,10 @@ class Pipeline:
             the learned transformations on the incoming df.
         return_lazy
             If true, return the lazy plan for the transformations
+        separate
+            Separate the feature data (usually referred to as `X`), from the target data, (
+            usually referred to as `y`.) by returning two dataframes representing features and
+            the target.
         """
         if self.ensure_features_in:
             if _IS_POLARS_V1:
@@ -352,7 +372,7 @@ class Pipeline:
             else:
                 extras = [c for c in df.columns if c not in self.feature_names_in_]
                 missing = [c for c in self.feature_names_in_ if c not in df.columns]
-            if len(extras) > 0 or len(missing):
+            if len(extras) > 0 or len(missing) > 0:
                 raise ValueError(
                     f"Input df doesn't have the features expected. Extra columns: {extras}. Missing columns: {missing}"
                 )
@@ -360,7 +380,24 @@ class Pipeline:
         plan = self._generate_lazy_plan(df)
         if self.ensure_features_out:
             plan = plan.select(self.feature_names_out_)
-        return plan if return_lazy else plan.collect()  # Add config here if streaming is needed
+        
+        # Add config here if streaming is needed
+        if separate:
+            if self.target is None:
+                raise ValueError("If you want to separate the feature data from target data, please specify the target at blueprint initiation.")
+            else:
+                if return_lazy:
+                    return (
+                        plan.select(pl.all().exclude(self.target)),
+                        plan.select(self.target)
+                    )
+                else:
+                    return (
+                        plan.select(pl.all().exclude(self.target)).collect(),
+                        plan.select(self.target).collect()
+                    )
+        else:
+            return plan if return_lazy else plan.collect()  
 
 
 class Blueprint:
@@ -416,7 +453,7 @@ class Blueprint:
         out += f"Features Expected: {self.feature_names_in_}\n"
         return out
 
-    def _get_target(self, target: str | pl.Expr | None = None) -> StrOrExpr:
+    def _get_target(self, target: str | pl.Expr | None = None) -> str | pl.Expr:
         if target is None:
             if self.target is None:
                 raise ValueError(
@@ -446,33 +483,39 @@ class Blueprint:
     #     self._steps = [deepcopy(s) for s in self._steps]
     #     return self
 
-    def filter(self, *by: str | pl.Expr, all_true: bool = True) -> Self:
+    def filter(self, by: str | pl.Expr) -> Self:
         """
-        Filters on the dataframe using native polars expressions or SQL strings (the part after where).
-        Note, elements in `by` must either all be strings or all be polars expressions.
+        Filters on the dataframe using native polars expressions or SQL boolean expressions.
 
         Parameters
         ----------
         by
             Native polars boolean expression or SQL strings
-        all_true
-            Whether all conditions should be met, or any.
         """
-        inputs = list(by)
-        is_all_string = all(isinstance(b, str) for b in inputs)
-        is_all_expr = all(isinstance(b, pl.Expr) for b in inputs)
-        if (is_all_string or is_all_expr) is False:  # if both are false
-            raise ValueError("Filters must either all be strings or all be polars expressions.")
+        self._steps.append(FilterStep(by if isinstance(by, pl.Expr) else pl.sql_expr(by)))
+        return self
 
-        if is_all_string:
-            exprs = [pl.sql_expr(s) for s in inputs]
-        else:  # is_all_expr
-            exprs = list(inputs)
+    def sql_transform(self, sql: str) -> Self:
+        """
+        Runs the SQL on the dataframe when it reaches this step. The user must ensure that
+        the SQL is valid Polars SQL and all columns referred in the SQL exist at this point.
+        The name "df" should be used to refer to the current state of the dataframe in the SQL.
+        E.g. select * from df where A is True.
 
-        if all_true:
-            self._steps.append(FilterStep(pl.all_horizontal(exprs)))
-        else:
-            self._steps.append(FilterStep(pl.any_horizontal(exprs)))
+        Parameters
+        ----------
+        sql
+            The SQL to run on the dataframe. Note: this step doesn't immedinately check the validity of
+            the SQL statement.
+        """
+        self._steps.append(SQLStep(sql))
+        return self
+
+    def cast_bools(self, to: pl.DataType = pl.UInt8) -> Self:
+        """
+        Cast all boolean columns in the dataframe to the given type.
+        """
+        self._steps.append(WithColumnsStep(cs.boolean().cast(to)))
         return self
 
     def impute(self, cols: IntoExprColumn, method: SimpleImputeMethod = "mean") -> Self:
@@ -491,9 +534,30 @@ class Blueprint:
         self._steps.append(FitStep(partial(t.impute, method=method), cols, self.exclude))
         return self
 
+    def conditional_impute(
+        self, 
+        rules_dict: Dict[str, str | pl.Expr],
+        method:SimpleImputeMethod = "mean"
+    ) -> Self:
+        """
+        Conditionally imputes values in the given columns. This transform will collect if input is lazy.
+
+        Parameters
+        ----------
+        rules_dict
+            Dictionary where keys are column names (must be string), and values are SQL/Polars Conditions 
+            that when true, those values in the column will be imputed, 
+            and the value to impute will be learned on the data where the condition is false.
+        method
+            One of `mean`, `median`, `mode`. If `mode`, a random value will be chosen if there is
+            a tie.
+        """
+        self._steps.append(FitStep(partial(t.conditional_impute, rules_dict=rules_dict, method=method), None, self.exclude))
+        return self
+
     def nan_to_null(self) -> Self:
         """
-        Maps NaN values to null.
+        Maps NaN values in all columns to null.
         """
         self._steps.append(WithColumnsStep(cs.float().nan_to_null()))
         return self
@@ -542,7 +606,7 @@ class Blueprint:
         self._steps.append(FitStep(partial(t.scale, method=method), cols, self.exclude))
         return self
 
-    def robust_scale(self, cols: IntoExprColumn, q1: float, q2: float) -> Self:
+    def robust_scale(self, cols: IntoExprColumn, q_low: float, q_high: float) -> Self:
         """
         Performs robust scaling on the given columns
 
@@ -550,12 +614,12 @@ class Blueprint:
         ---------
         cols
             Any Polars expression that can be understood as columns.
-        q1
+        q_low
             The lower quantile value
-        q2
+        q_high
             The higher quantile value
         """
-        self._steps.append(FitStep(partial(t.robust_scale, q1=q1, q2=q2), cols, self.exclude))
+        self._steps.append(FitStep(partial(t.robust_scale, q_low=q_low, q_high=q_high), cols, self.exclude))
         return self
 
     def center(self, cols: IntoExprColumn) -> Self:
@@ -582,7 +646,7 @@ class Blueprint:
         self._steps.append(SelectStep(cols))
         return self
 
-    def shrink_dtype(self, force_f32: bool = False) -> pl.Expr:
+    def shrink_dtype(self, force_f32: bool = False) -> Self:
         """
         Shrinks the dtype by calling shrink_dtype on all numerical columns. This may reduce
         the memory pressure during the process.
@@ -590,7 +654,8 @@ class Blueprint:
         Parameters
         ----------
         force_f32
-            If true, force all float columns to be f32 type.
+            If true, force all float columns to be f32 type. You might want to consider using f32
+            only in the pipeline if you wish to save memory.
         """
 
         exprs = cs.integer().shrink_dtype()
@@ -600,7 +665,9 @@ class Blueprint:
 
         return self
 
-    def polynomial_features(self, cols: List[str], degree: int, interaction_only: bool = True):
+    def polynomial_features(
+        self, cols: List[str], degree: int, interaction_only: bool = True
+    ) -> Self:
         """
         Generates polynomial combinations out of the features given, at the given degree.
 
@@ -628,8 +695,8 @@ class Blueprint:
     def winsorize(
         self,
         cols: IntoExprColumn,
-        lower: float = 0.05,
-        upper: float = 0.95,
+        q_low: float = 0.05,
+        q_high: float = 0.95,
         method: QuantileMethod = "nearest",
     ) -> Self:
         """
@@ -641,16 +708,16 @@ class Blueprint:
         ----------
         cols
             Any Polars expression that can be understood as columns. Columns must be numerical.
-        lower
+        q_low
             The lower quantile value
-        upper
+        q_high
             The higher quantile value
         method
             Method to compute quantile. One of `nearest`, `higher`, `lower`, `midpoint`, `linear`.
         """
         self._steps.append(
             FitStep(
-                partial(t.winsorize, lower=lower, upper=upper, method=method),
+                partial(t.winsorize, q_low=q_low, q_high=q_high, method=method),
                 cols,
                 self.exclude,
             )
@@ -669,7 +736,7 @@ class Blueprint:
         self._steps.append(SelectStep(pl.all().exclude(cols)))
         return self
 
-    def rename(self, rename_dict: Dict[str, str]) -> pl.Expr:
+    def rename(self, rename_dict: Dict[str, str]) -> Self:
         """
         Renames the columns by the mapping.
 
@@ -694,6 +761,18 @@ class Blueprint:
             self._steps.append(SelectStep([pl.col(c).alias(c.lower()) for c in self._df.columns]))
         return self
 
+    def uppercase(self) -> Self:
+        """
+        Uppercases all column names.
+        """
+        if _IS_POLARS_V1:
+            self._steps.append(
+                SelectStep([pl.col(c).alias(c.upper()) for c in self._df.collect_schema().names()])
+            )
+        else:
+            self._steps.append(SelectStep([pl.col(c).alias(c.upper()) for c in self._df.columns]))
+        return self
+
     def one_hot_encode(
         self,
         cols: IntoExprColumn,
@@ -968,21 +1047,24 @@ class Blueprint:
         # the collect should be and optimized.
         df_lazy: pl.LazyFrame = df.lazy()
         for step in self._steps:
-            if isinstance(step, FitStep):
+            if isinstance(step, FitStep):  # Need fitting
                 exprs = step.fit(df_lazy)
                 transforms.append(WithColumnsStep(exprs))
                 df_lazy = df_lazy.with_columns(exprs)
-            elif isinstance(step, WithColumnsStep):
+            elif isinstance(step, WithColumnsStep):  # Fitted
                 transforms.append(step)
                 df_lazy = df_lazy.with_columns(step.exprs)
-            elif isinstance(step, SelectStep):
+            elif isinstance(step, SelectStep):  # Fitted
                 transforms.append(step)
                 df_lazy = df_lazy.select(step.exprs)
-            elif isinstance(step, FilterStep):
+            elif isinstance(step, FilterStep):  # Fitted
                 transforms.append(step)
                 df_lazy = df_lazy.filter(step.exprs)
+            elif isinstance(step, SQLStep):  # Fitted
+                transforms.append(step)
+                df_lazy = pl.SQLContext(df=df_lazy).execute(step.sql)
             else:
-                raise ValueError("Not a valid step.")
+                raise ValueError(f"Not a valid step: {step.__class__}")
 
         return Pipeline(
             name=self.name,
@@ -1000,11 +1082,20 @@ class Blueprint:
         """
         return self.materialize()
 
-    def transform(self, df: PolarsFrame) -> pl.DataFrame:
-        return self.materialize().transform(df)
+    def transform(self, 
+        df: PolarsFrame,
+        separate: bool = False
+    ) -> pl.DataFrame | Tuple[pl.DataFrame, pl.DataFrame]:
+        """
+        Fits the blueprint with the dataframe that it is initialized with, and 
+        transforms the input dataframe.
 
-    def finish(self) -> Pipeline:
+        Parameters
+        ----------
+        df
+            Any Polars dataframe
+        separate
+            If true, two dataframes will be returned, one representing the feature
+            data and the other the target data. If false, then only one dataframe is returned.
         """
-        Alias for self.materialize()
-        """
-        return self.materialize()
+        return self.materialize().transform(df, separate=separate)
