@@ -1,19 +1,20 @@
+"""Simple Statistics."""
+
 from __future__ import annotations
 
 import polars as pl
 import math
-from .type_alias import Alternative, str_to_expr, CorrMethod, Noise, QuantileMethod
-from typing import Union
-from ._utils import pl_plugin
+from .typing import Alternative, CorrMethod, Noise, QuantileMethod
+from ._utils import pl_plugin, str_to_expr
 
 __all__ = [
-    "query_ttest_ind",
-    "query_ttest_1samp",
-    "query_ttest_ind_from_stats",
-    "query_ks_2samp",
-    "query_f_test",
-    "query_mann_whitney_u",
-    "query_chi2",
+    "ttest_ind",
+    "ttest_1samp",
+    "ttest_ind_from_stats",
+    "ks_2samp",
+    "f_test",
+    "mann_whitney_u",
+    "chi2",
     "perturb",
     "jitter",
     "add_noise",
@@ -41,7 +42,7 @@ __all__ = [
 ]
 
 
-def query_ttest_ind(
+def ttest_ind(
     var1: str | pl.Expr,
     var2: str | pl.Expr,
     alternative: Alternative = "two-sided",
@@ -99,7 +100,7 @@ def query_ttest_ind(
         )
 
 
-def query_ttest_1samp(
+def ttest_1samp(
     var1: str | pl.Expr, pop_mean: float, alternative: Alternative = "two-sided"
 ) -> pl.Expr:
     """
@@ -129,7 +130,7 @@ def query_ttest_1samp(
     )
 
 
-def query_ttest_ind_from_stats(
+def ttest_ind_from_stats(
     var1: str | pl.Expr,
     mean: float,
     var: float,
@@ -185,7 +186,7 @@ def query_ttest_ind_from_stats(
         )
 
 
-def query_ks_2samp(
+def ks_2samp(
     var1: str | pl.Expr,
     var2: str | pl.Expr,
     alpha: float = 0.05,
@@ -201,7 +202,7 @@ def query_ks_2samp(
     hypothesis should be rejected. This is suitable only for large sameple sizes. See more details
     in the reference.
 
-    If either var1 or var2 has less than 20 values, a ks stats of INFINITY will be returned.
+    If either var1 or var2 has less than 30 values, a ks stats of 0 with threshold NaN will be returned.
 
     Parameters
     ----------
@@ -220,25 +221,22 @@ def query_ks_2samp(
     """
     y1, y2 = str_to_expr(var1), str_to_expr(var2)
     if is_binary:
-        z = y2.filter(y2.is_finite()).cast(pl.Float64)
-        z1 = z.filter(y1 == 1).sort()
-        z2 = z.filter(y1 == 0).sort()
-        return pl_plugin(
-            symbol="pl_ks_2samp",
-            args=[z1, z2, pl.lit(alpha, pl.Float64)],
-            returns_scalar=True,
-        )
+        z1 = y2.filter(y1 == 1)
+        z2 = y2.filter(y1 == 0)
+        z1 = z1.filter(z1.is_finite()).sort()
+        z2 = z2.filter(z2.is_finite()).sort()
     else:
         z1 = y1.filter(y1.is_finite()).sort()
         z2 = y2.filter(y2.is_finite()).sort()
-        return pl_plugin(
-            symbol="pl_ks_2samp",
-            args=[z1, z2, pl.lit(alpha, pl.Float64)],
-            returns_scalar=True,
-        )
+
+    return pl_plugin(
+        symbol="pl_ks_2samp",
+        args=[z1, z2, pl.lit(alpha, pl.Float64)],
+        returns_scalar=True,
+    )
 
 
-def query_f_test(*variables: str | pl.Expr, group: str | pl.Expr) -> pl.Expr:
+def f_test(*variables: str | pl.Expr, group: str | pl.Expr) -> pl.Expr:
     """
     Performs the ANOVA F-test.
 
@@ -259,7 +257,7 @@ def query_f_test(*variables: str | pl.Expr, group: str | pl.Expr) -> pl.Expr:
         return pl_plugin(symbol="pl_f_test", args=vars_, changes_length=True)
 
 
-def query_chi2(var1: str | pl.Expr, var2: str | pl.Expr) -> pl.Expr:
+def chi2(var1: str | pl.Expr, var2: str | pl.Expr) -> pl.Expr:
     """
     Computes the Chi Squared statistic and p value between two categorical values.
 
@@ -281,7 +279,7 @@ def query_chi2(var1: str | pl.Expr, var2: str | pl.Expr) -> pl.Expr:
     )
 
 
-def query_mann_whitney_u(
+def mann_whitney_u(
     var1: str | pl.Expr,
     var2: str | pl.Expr,
     alternative: Alternative = "two-sided",
@@ -327,8 +325,8 @@ def query_mann_whitney_u(
 
 def winsorize(
     x: str | pl.Expr,
-    lower: float = 0.05,
-    upper: float = 0.95,
+    q_low: float = 0.05,
+    q_high: float = 0.95,
     method: QuantileMethod = "nearest",
 ) -> pl.Expr:
     """
@@ -338,25 +336,27 @@ def winsorize(
     ----------
     x
         Either the name of the column or a Polars expression
-    lower
+    q_low
         The lower percentile value to clip the data. E.g everything < x.quantile(lower)
         will be mapped to x.quantile(lower)
-    upper
+    q_high
         The upper percentile value to clip the data. E.g everything > x.quantile(upper)
         will be mapped to x.quantile(upper)
     method
         Method for quantile estimate. One of "nearest", "higher", "lower", "midpoint", "linear".
     """
-    if lower <= 0.0 or lower >= 1.0 or upper <= 0.0 or upper >= 1.0 or upper <= lower:
+    if q_low <= 0.0 or q_low >= 1.0 or q_high <= 0.0 or q_high >= 1.0 or q_high <= q_low:
         raise ValueError("Lower and upper must be with in (0, 1) and upper should be > lower")
 
     xx = str_to_expr(x)
     return xx.clip(
-        xx.quantile(lower, interpolation=method), xx.quantile(upper, interpolation=method)
+        xx.quantile(q_low, interpolation=method), xx.quantile(q_high, interpolation=method)
     )
 
 
-def perturb(x: str | pl.Expr, epsilon: float, positive: bool = False):
+def perturb(
+    x: str | pl.Expr, epsilon: float = 1e-5, positive: bool = False, seed: int | None = None
+) -> pl.Expr:
     """
     Perturb the var by a small amount. This only applies to float columns.
 
@@ -369,6 +369,8 @@ def perturb(x: str | pl.Expr, epsilon: float, positive: bool = False):
     positive
         If true, randomly add a small amount in [0, epsilon). If false, it will use the range
         [-epsilon/2, epsilon/2)
+    seed
+        A random seed
     """
     if math.isinf(epsilon) or math.isnan(epsilon):
         raise ValueError("Input `epsilon should be a valid finite value.`")
@@ -384,12 +386,12 @@ def perturb(x: str | pl.Expr, epsilon: float, positive: bool = False):
 
     return pl_plugin(
         symbol="pl_perturb",
-        args=[str_to_expr(x), lo, hi],
+        args=[str_to_expr(x), lo, hi, pl.lit(seed, dtype=pl.UInt64)],
         is_elementwise=True,
     )
 
 
-def jitter(x: str | pl.Expr, std: float | pl.Expr = 1.0) -> pl.Expr:
+def jitter(x: str | pl.Expr, std: float | pl.Expr = 1.0, seed: int | None = None) -> pl.Expr:
     """
     Adds a Gaussian noise of N(0, std) to the column.
 
@@ -399,6 +401,8 @@ def jitter(x: str | pl.Expr, std: float | pl.Expr = 1.0) -> pl.Expr:
         Either the name of the column or a Polars expression
     std
         The std of the Gaussian noise.
+    seed
+        A random seed
     """
     if isinstance(std, float):
         if std < 0:
@@ -412,7 +416,7 @@ def jitter(x: str | pl.Expr, std: float | pl.Expr = 1.0) -> pl.Expr:
 
     return pl_plugin(
         symbol="pl_jitter",
-        args=[str_to_expr(x), s],
+        args=[str_to_expr(x), s, pl.lit(seed, dtype=pl.UInt64)],
         is_elementwise=True,
     )
 
@@ -429,7 +433,7 @@ def add_noise(x: str | pl.Expr, noise_type: Noise = "gaussian", **kwargs) -> pl.
         Either "gaussian" or "uniform"
     kwargs
         If noise_type = "gaussian", this accepts kwargs to "jitter" and if "uniform", this
-        accepts kwargs to "perturb".
+        accepts kwargs to "perturb". You may set a seed via the kwargs.
     """
     if noise_type == "gaussian":
         return jitter(x, **kwargs)
@@ -470,9 +474,10 @@ def normal_test(var: str | pl.Expr) -> pl.Expr:
 
 
 def random(
-    lower: Union[pl.Expr, float] = 0.0,
-    upper: Union[pl.Expr, float] = 1.0,
+    lower: pl.Expr | float = 0.0,
+    upper: pl.Expr | float = 1.0,
     seed: int | None = None,
+    length: int | pl.Expr = pl.len(),
 ) -> pl.Expr:
     """
     Generate random numbers in [lower, upper)
@@ -485,24 +490,27 @@ def random(
         The upper bound, exclusive
     seed
         The random seed. None means no seed.
+    length
+        Custom length. Note length needs to match with other columns in the context.
     """
     lo = pl.lit(lower, pl.Float64) if isinstance(lower, float) else lower
     up = pl.lit(upper, pl.Float64) if isinstance(upper, float) else upper
+    len_ = pl.lit(length, pl.UInt32) if isinstance(length, int) else length
     return pl_plugin(
         symbol="pl_random",
-        args=[pl.len(), lo, up, pl.lit(seed, pl.UInt64)],
+        args=[len_, lo, up, pl.lit(seed, pl.UInt64)],
         is_elementwise=True,
     )
 
 
-def random_null(var: str | pl.Expr, pct: float, seed: int | None = None) -> pl.Expr:
+def random_null(x: str | pl.Expr, pct: float, seed: int | None = None) -> pl.Expr:
     """
-    Creates random null values in var. If var contains nulls originally, they
+    Creates random null values in the columns. If var contains nulls originally, they
     will stay null.
 
     Parameters
     ----------
-    var
+    x
         Either the name of the column or a Polars expression
     pct
         Percentage of nulls to randomly generate. This percentage is based on the
@@ -514,12 +522,14 @@ def random_null(var: str | pl.Expr, pct: float, seed: int | None = None) -> pl.E
     if pct <= 0.0 or pct >= 1.0:
         raise ValueError("Input `pct` must be > 0 and < 1")
 
-    to_null = random(0.0, 1.0, seed=seed) < pct
-    return pl.when(to_null).then(None).otherwise(str_to_expr(var))
+    return pl.when(random(0.0, 1.0, seed=seed) < pct).then(None).otherwise(str_to_expr(x))
 
 
 def random_int(
-    lower: Union[int, pl.Expr], upper: Union[int, pl.Expr], seed: int | None = None
+    lower: int | pl.Expr, 
+    upper: int | pl.Expr, 
+    seed: int | None = None,
+    length: int | pl.Expr = pl.len(),
 ) -> pl.Expr:
     """
     Generates random integer between lower and upper.
@@ -532,16 +542,19 @@ def random_int(
         The upper bound, exclusive
     seed
         The random seed. None means no seed.
+    length
+        Custom length. Note length needs to match with other columns in the context.
     """
     if lower == upper:
         raise ValueError("Input `lower` must be smaller than `higher`")
 
     lo = pl.lit(lower, pl.Int32) if isinstance(lower, int) else lower.cast(pl.Int32)
     hi = pl.lit(upper, pl.Int32) if isinstance(upper, int) else upper.cast(pl.Int32)
+    len_ = pl.lit(length, pl.UInt32) if isinstance(length, int) else length
     return pl_plugin(
         symbol="pl_rand_int",
         args=[
-            pl.len().cast(pl.UInt32),
+            len_,
             lo,
             hi,
             pl.lit(seed, pl.UInt64),
@@ -550,9 +563,15 @@ def random_int(
     )
 
 
-def random_str(min_size: int, max_size: int) -> pl.Expr:
+def random_str(
+    min_size: int, 
+    max_size: int, 
+    seed: int | None = None,
+    length: int | pl.Expr = pl.len(),
+) -> pl.Expr:
     """
-    Generates random strings of length between min_size and max_size.
+    Generates random strings of length between min_size and max_size. The characters are
+    uniformly distributed over ASCII letters and numbers: a-z, A-Z and 0-9.
 
     Parameters
     ----------
@@ -562,6 +581,8 @@ def random_str(min_size: int, max_size: int) -> pl.Expr:
         The max size of the string, inclusive
     seed
         The random seed. None means no seed.
+    length
+        Custom length. Note length needs to match with other columns in the context.
     """
     mi, ma = min_size, max_size
     if min_size > max_size:
@@ -570,16 +591,16 @@ def random_str(min_size: int, max_size: int) -> pl.Expr:
     return pl_plugin(
         symbol="pl_rand_str",
         args=[
-            pl.len().cast(pl.UInt32),
+            pl.lit(length, pl.UInt32) if isinstance(length, int) else length,
             pl.lit(mi, pl.UInt32),
             pl.lit(ma, pl.UInt32),
-            pl.lit(42, pl.UInt64),
+            pl.lit(seed, pl.UInt64),
         ],
         is_elementwise=True,
     )
 
 
-def random_binomial(n: int, p: int, seed: int | None = None) -> pl.Expr:
+def random_binomial(n: int, p: int, seed: int | None = None, length: int | pl.Expr = pl.len()) -> pl.Expr:
     """
     Generates random integer following a binomial distribution.
 
@@ -591,6 +612,8 @@ def random_binomial(n: int, p: int, seed: int | None = None) -> pl.Expr:
         The p in a binomial distribution
     seed
         The random seed. None means no seed.
+    length
+        Custom length. Note length needs to match with other columns in the context.
     """
     if n < 1:
         raise ValueError("Input `n` must be > 1.")
@@ -598,7 +621,7 @@ def random_binomial(n: int, p: int, seed: int | None = None) -> pl.Expr:
     return pl_plugin(
         symbol="pl_rand_binomial",
         args=[
-            pl.len().cast(pl.UInt32),
+            pl.lit(length, pl.UInt32) if isinstance(length, int) else length,
             pl.lit(n, pl.Int32),
             pl.lit(p, pl.Float64),
             pl.lit(seed, pl.UInt64),
@@ -607,7 +630,7 @@ def random_binomial(n: int, p: int, seed: int | None = None) -> pl.Expr:
     )
 
 
-def random_exp(lambda_: float, seed: int | None = None) -> pl.Expr:
+def random_exp(lambda_: float, seed: int | None = None, length: int | pl.Expr = pl.len()) -> pl.Expr:
     """
     Generates random numbers following an exponential distribution.
 
@@ -617,16 +640,25 @@ def random_exp(lambda_: float, seed: int | None = None) -> pl.Expr:
         The lambda in an exponential distribution
     seed
         The random seed. None means no seed.
+    length
+        Custom length. Note length needs to match with other columns in the context.
     """
     return pl_plugin(
         symbol="pl_rand_exp",
-        args=[pl.len().cast(pl.UInt32), pl.lit(lambda_, pl.Float64), pl.lit(seed, pl.UInt64)],
+        args=[
+            pl.lit(length, pl.UInt32) if isinstance(length, int) else length,
+            pl.lit(lambda_, pl.Float64), 
+            pl.lit(seed, pl.UInt64)
+        ],
         is_elementwise=True,
     )
 
 
 def random_normal(
-    mean: Union[pl.Expr, float], std: Union[pl.Expr, float], seed: int | None = None
+    mean: pl.Expr | float, 
+    std: pl.Expr | float, 
+    seed: int | None = None,
+    length: int | pl.Expr = pl.len()
 ) -> pl.Expr:
     """
     Generates random number following a normal distribution.
@@ -639,12 +671,17 @@ def random_normal(
         The std in a normal distribution
     seed
         The random seed. None means no seed.
+    length
+        Custom length. Note length needs to match with other columns in the context.
     """
-    m = pl.lit(mean, pl.Float64) if isinstance(mean, float) else mean
-    s = pl.lit(std, pl.Float64) if isinstance(std, float) else std
     return pl_plugin(
         symbol="pl_rand_normal",
-        args=[pl.len().cast(pl.UInt32), m, s, pl.lit(seed, pl.UInt64)],
+        args=[
+            pl.lit(length, pl.UInt32) if isinstance(length, int) else length, 
+            pl.lit(mean, pl.Float64) if isinstance(mean, float) else mean, 
+            pl.lit(std, pl.Float64) if isinstance(std, float) else std, 
+            pl.lit(seed, pl.UInt64)
+        ],
         is_elementwise=True,
     )
 
@@ -750,7 +787,7 @@ def weighted_var(var: str | pl.Expr, weights: str | pl.Expr, freq_weights: bool
     return summand / w.sum()
 
 
-def weighted_cov(x: str | pl.Expr, y: str | pl.Expr, weights: Union[pl.Expr, float]) -> pl.Expr:
+def weighted_cov(x: str | pl.Expr, y: str | pl.Expr, weights: pl.Expr | float) -> pl.Expr:
     """
     Computes the weighted covariance between x and y. The weights column must have the same
     length as both x an y.
