@@ -1,9 +1,10 @@
+"""String-related utils, including edit distances, simple cleaning, etc."""
+
 from __future__ import annotations
 
-from typing import List, Literal, Dict
 import polars as pl
-from ._utils import pl_plugin
-from .type_alias import str_to_expr
+from typing import List, Literal, Dict
+from ._utils import pl_plugin, str_to_expr
 
 
 __all__ = [
@@ -15,7 +16,7 @@ __all__ = [
     "to_snake_case",
     "to_pascal_case",
     "to_constant_case",
-    "query_similar_words",
+    "str_nearest",
     "str_snowball",
     "str_tokenize",
     "str_jaccard",
@@ -26,6 +27,7 @@ __all__ = [
     "str_d_leven",
     "str_leven",
     "str_osa",
+    "str_lcs_seq",
     "str_fuzz",
     "similar_to_vocab",
     "extract_numbers",
@@ -189,70 +191,45 @@ def to_constant_case(c: str | pl.Expr) -> pl.Expr:
     )
 
 
-def query_similar_words(
+def str_nearest(
     c: str | pl.Expr,
-    vocab: pl.Expr | List[str],
-    k: int = 1,
+    word: str,
     threshold: int = 100,
     metric: Literal["lv", "hamming"] = "lv",
-    parallel: bool = False,
 ) -> pl.Expr:
     """
-    Finds the k most similar words in vocab to each word in self. This works by computing
-    Hamming/Levenshtein distances, instead of similarity, and then taking the smallest distance
-    words as similar words. The result is correct because of the relationship between distance
-    and similarity. This will deduplicate words in vocab. In case of a tie, any one may be chosen.
+    Finds the string in the column that is nearest to the given word in the given metric.
 
-    Comment: This can be very slow due to underlying data structure problems. Setting a threshold
-    may speed up the process a little bit.
+    Note: Nearest-k strings search functionality is temporarily dropped.
 
     Parameters
     ----------
-    vocab
+    c
+        The string column or its name
+    word
         Any iterable collection of strings that can be turned into a polars Series, or an expression
-    k : int, >0
-        k most similar words will be found
     threshold : int, >0
-        Only considers words to be similar if they are within distance threshold. This is a positive integer
+        Only considers strings to be near if they are within distance threshold. This is a positive integer
         because all the distances output integers.
     metric
         Which similarity metric to use. One of `lv`, `hamming`
-    parallel
-        Whether to run the comparisons in parallel. Note that this is only recommended when this query
-        is the only one in the context and we are not in any aggregation context.
     """
     if metric not in ("lv", "hamming"):
         raise ValueError(f"Unknown metric for similar_words: {metric}")
 
-    if isinstance(vocab, pl.Expr):
-        vb = vocab.unique().drop_nulls()
-    else:
-        vb = pl.Series(values=vocab, dtype=pl.String).unique().drop_nulls()
+    if threshold <= 0:
+        raise ValueError("Distance threshold must be > 0.")
 
-    if k == 1:  # k = 1, this is a fast path (no heap)
-        return pl_plugin(
-            symbol="pl_nearest_str",
-            args=[str_to_expr(c), vb],
-            kwargs={
-                "k": k,
-                "metric": str(metric).lower(),
-                "threshold": threshold,
-                "parallel": parallel,
-            },
-        )
-    elif k > 1:
-        return pl_plugin(
-            symbol="pl_knn_str",
-            args=[str_to_expr(c), vb],
-            kwargs={
-                "k": k,
-                "metric": str(metric).lower(),
-                "threshold": threshold,
-                "parallel": parallel,
-            },
-        )
-    else:
-        raise ValueError("Input `k` must be >= 1.")
+    return pl_plugin(
+        symbol="pl_nearest_str",
+        args=[str_to_expr(c)],
+        kwargs={
+            "word": word,
+            "metric": str(metric).lower(),
+            "threshold": threshold,
+        },
+        returns_scalar=True,
+    )
 
 
 def str_snowball(c: str | pl.Expr, no_stopwords: bool = True) -> pl.Expr:
@@ -590,6 +567,42 @@ def str_leven(
         )
 
 
+def str_lcs_seq(
+    c: str | pl.Expr,
+    other: str | pl.Expr,
+    parallel: bool = False,
+    return_sim: bool = True,
+) -> pl.Expr:
+    """
+    Computes the Longest Common Subsequence distance/similarity between this and the other str.
+    The distance is calculated as max(len1, len2) - similarity, where the similarity is the
+    the length of the longest common subsequence. Subsequences may not occupy consecutive positions.
+
+    Parameters
+    ----------
+    c
+        The string column
+    other
+        Either the name of the column or a Polars expression. If you want to compare a single
+        string with all of column c, use pl.lit(your_str)
+    parallel
+        Whether to run the comparisons in parallel. Note that this is only recommended when this query
+        is the only one in the context and we are not in any aggregation context.
+    return_sim
+        If true, return normalized similarity.
+    """
+    if return_sim:
+        return pl_plugin(
+            symbol="pl_lcs_seq_sim",
+            args=[str_to_expr(c), str_to_expr(other), pl.lit(parallel, pl.Boolean)],
+        )
+    else:
+        return pl_plugin(
+            symbol="pl_lcs_seq",
+            args=[str_to_expr(c), str_to_expr(other), pl.lit(parallel, pl.Boolean)],
+        )
+
+
 def str_osa(
     c: str | pl.Expr,
     other: str | pl.Expr,
@@ -626,7 +639,8 @@ def str_osa(
 
 def str_fuzz(c: str | pl.Expr, other: str | pl.Expr, parallel: bool = False) -> pl.Expr:
     """
-    A string similarity based on Longest Common Subsequence.
+    Calculates the normalized Indel similarity. (See the package rapidfuzz, fuzz.ratio for more
+    information.)
 
     Parameters
     ----------
@@ -653,8 +667,8 @@ def similar_to_vocab(
     strategy: Literal["avg", "all", "any"] = "avg",
 ) -> pl.Expr:
     """
-    Compare each word in the vocab with each word in self. Filters c to the words
-    that are most similar to the words in the vocab.
+    Compare each word in the vocab with each word in the column c. Returns a boolean
+    that indicates whether there exist words in c that are similar to words in vocab.
 
     Parameters
     ----------
@@ -912,7 +926,7 @@ def map_words(c: str | pl.Expr, mapping: Dict[str, str]) -> pl.Expr:
     c : str | pl.Expr
         The string column
     mapping : dict[str, str]
-        A dictionary of {word: replace_with}
+        A dictionary of {word: the replacement}
 
     Returns
     -------
