@@ -3,11 +3,12 @@ This module provides classic ML dataset transforms. Note all functions here are
 that the data learned (e.g. mean value in mean imputation) will not be preserved. For pipeline usage, which
 preserves the learned values and optimizes the transform query, see pipeline.py.
 """
+
 from __future__ import annotations
 
 import polars as pl
 import polars.selectors as cs
-from .type_alias import (
+from .typing import (
     PolarsFrame,
     SimpleImputeMethod,
     SimpleScaleMethod,
@@ -16,7 +17,7 @@ from .type_alias import (
     EncoderDefaultStrategy,
 )
 from . import num as pds_num
-from . import query_linear as pds_linear
+from . import expr_linear as lr
 from ._utils import _IS_POLARS_V1
 from typing import List
 
@@ -45,8 +46,47 @@ def impute(df: PolarsFrame, cols: List[str], method: SimpleImputeMethod = "mean"
         temp = df.lazy().select(pl.col(cols).mode().list.first()).collect().row(0)
         return [pl.col(c).fill_null(m) for c, m in zip(cols, temp)]
     else:
-        raise ValueError(f"Unknown input method: {method}")
+        raise ValueError(f"Unknown impute method: `{method}`")
 
+def conditional_impute(
+    df: PolarsFrame, 
+    rules_dict: Dict[str, str | pl.Expr],
+    method: SimpleImputeMethod = "mean"
+) -> ExprTransform:
+    """
+    Conditionally imputes values in the given columns. This transform will collect if input is lazy.
+
+    Parameters
+    ----------
+    df
+        Either a lazy or an eager dataframe
+    rules_dict
+        Dictionary where keys are column names (must be string), and values are SQL/Polars Conditions 
+        that when true, those values in the column will be imputed, 
+        and the value to impute will be learned on the data where the condition is false.
+    method
+        One of `mean`, `median`, `mode`. If `mode`, a random value will be chosen if there is
+        a tie.
+    """
+    rules_dict = {
+        c: (r if isinstance(r, pl.Expr) else pl.sql_expr(r))
+        for c, r in rules_dict.items()
+    }
+    cols = list(rules_dict.keys())
+    # Learn on the data where the condition is false
+    if method == "mean":
+        temp = df.lazy().select(
+            *(pl.col(c).filter(rules_dict[c].not_()).mean() for c in rules_dict.keys())
+        ).collect().row(0)
+        return [pl.when(rules_dict[c]).then(m).otherwise(pl.col(c)).alias(c) for c, m in zip(cols, temp)]
+    elif method == "median":
+        temp = df.lazy().select(*(pl.col(c).filter(rules_dict[c].not_()).median() for c in rules_dict.keys())).collect().row(0)
+        return [pl.when(rules_dict[c]).then(m).otherwise(pl.col(c)).alias(c) for c, m in zip(cols, temp)]
+    elif method == "mode": 
+        temp = df.lazy().select(*(pl.col(c).filter(rules_dict[c].not_()).mode().list.first() for c in rules_dict.keys())).collect().row(0)
+        return [pl.when(rules_dict[c]).then(m).otherwise(pl.col(c)).alias(c) for c, m in zip(cols, temp)]
+    else:
+        raise ValueError(f"Unknown impute method: `{method}`")
 
 def linear_impute(
     df: PolarsFrame, features: List[str], target: str | pl.Expr, add_bias: bool = False
@@ -81,7 +121,7 @@ def linear_impute(
     temp = (
         df.lazy()
         .select(
-            pds_linear.query_lstsq(
+            lr.lin_reg(
                 *features_as_expr, target=target_as_expr, add_bias=add_bias, null_policy="skip"
             )
         )
@@ -172,8 +212,8 @@ def scale(
 def robust_scale(
     df: PolarsFrame,
     cols: List[str],
-    q1: float = 0.25,
-    q2: float = 0.75,
+    q_low: float = 0.25,
+    q_high: float = 0.75,
     method: QuantileMethod = "midpoint",
 ) -> ExprTransform:
     """
@@ -185,21 +225,21 @@ def robust_scale(
         Either a lazy or an eager dataframe
     cols
         A list of strings representing column names
-    q1
+    q_low
         The lower quantile value
-    q2
+    q_high
         The higher quantile value
     method
         Method to compute quantile. One of `nearest`, `higher`, `lower`, `midpoint`, `linear`.
     """
-    if q1 > 1.0 or q1 < 0.0 or q2 > 1.0 or q2 < 0.0 or q1 >= q2:
-        raise ValueError("Input `q1` and `q2` must be between 0 and 1 and q1 must be < than q2.")
+    if q_low > 1.0 or q_low < 0.0 or q_high > 1.0 or q_high < 0.0 or q_low >= q_high:
+        raise ValueError("Input `q_low` and `q_high` must be between 0 and 1 and q_low must be < than q_high.")
 
     temp = (
         df.lazy()
         .select(
-            pl.col(cols).quantile(q1).name.prefix("q1:"),
-            pl.col(cols).quantile(q2).name.prefix("q2:"),
+            pl.col(cols).quantile(q_low, interpolation=method).name.prefix("q1:"),
+            pl.col(cols).quantile(q_high, interpolation=method).name.prefix("q2:"),
         )
         .collect()
         .row(0)
@@ -211,8 +251,8 @@ def robust_scale(
 def winsorize(
     df: PolarsFrame,
     cols: List[str],
-    lower: float = 0.05,
-    upper: float = 0.95,
+    q_low: float = 0.05,
+    q_high: float = 0.95,
     method: QuantileMethod = "nearest",
 ) -> ExprTransform:
     """
@@ -224,21 +264,21 @@ def winsorize(
         Either a lazy or an eager dataframe
     cols
         A list of strings representing column names
-    lower
+    q_low
         The lower quantile value
-    upper
+    q_high
         The higher quantile value
     method
         Method to compute quantile. One of `nearest`, `higher`, `lower`, `midpoint`, `linear`.
     """
-    if lower >= 1.0 or lower <= 0.0 or upper >= 1.0 or upper <= 0.0 or lower >= upper:
-        raise ValueError("Lower and upper must be with in (0, 1) and upper should be > lower")
+    if q_low > 1.0 or q_low < 0.0 or q_high > 1.0 or q_high < 0.0 or q_low >= q_high:
+        raise ValueError("Input `q_low` and `q_high` must be between 0 and 1 and q_low must be < than q_high.")
 
     temp = (
         df.lazy()
         .select(
-            pl.col(cols).quantile(lower).name.prefix("l"),
-            pl.col(cols).quantile(upper).name.prefix("u"),
+            pl.col(cols).quantile(q_low, interpolation=method).name.prefix("l"),
+            pl.col(cols).quantile(q_high, interpolation=method).name.prefix("u"),
         )
         .collect()
         .row(0)
@@ -489,7 +529,7 @@ def woe_encode(
     default_value = _encoder_default_value(temp, default=default, target=target)
 
     temp = temp.select(
-        pds_num.query_woe_discrete(c, target).implode() for c in valid_cols
+        pds_num.woe_discrete(c, target).implode() for c in valid_cols
     ).collect()  # add collect config..
     # POLARS_V1
     if _IS_POLARS_V1:
@@ -558,7 +598,7 @@ def iv_encode(
     default_value = _encoder_default_value(temp, default=default, target=target)
 
     temp = temp.select(
-        pds_num.query_iv_discrete(c, target, return_sum=False).implode() for c in valid_cols
+        pds_num.info_value_discrete(c, target, return_sum=False).implode() for c in valid_cols
     ).collect()  # add collect config..
     # POLARS_V1
     if _IS_POLARS_V1:
