@@ -1,8 +1,7 @@
 #![allow(non_snake_case)]
 use super::LinalgErrors;
+use core::f64;
 use faer::{prelude::*, Side};
-use faer_ext::IntoFaer;
-use ndarray::ArrayView2;
 use std::ops::Neg;
 
 #[derive(Clone, Copy, Default)]
@@ -31,6 +30,7 @@ pub enum LRMethods {
     Normal, // Normal. Normal Equation
     L1, // Lasso, L1 regularized
     L2, // Ridge, L2 regularized
+    ElasticNet,
 }
 
 impl From<&str> for LRMethods {
@@ -38,11 +38,29 @@ impl From<&str> for LRMethods {
         match value {
             "l1" | "lasso" => Self::L1,
             "l2" | "ridge" => Self::L2,
+            "elastic" => Self::ElasticNet,
             _ => Self::Normal,
         }
     }
 }
 
+/// Converts a 2-tuple of floats into LRMethods
+/// The first entry is assumed to the l1 regularization factor, and
+/// the second is assumed to be the l2 regularization factor
+impl From<(f64, f64)> for LRMethods {
+    fn from(value: (f64, f64)) -> Self {
+        if value.0 > 0. && value.1 <= 0. {
+            LRMethods::L1
+        } else if value.0 <= 0. && value.1 > 0. {
+            LRMethods::L2
+        } else if value.0 > 0. && value.1 > 0. {
+            LRMethods::ElasticNet
+        } else {
+            LRMethods::Normal
+        }
+    }
+}
+
 // add elastic net
 #[derive(Clone, Copy, Default, PartialEq)]
 pub enum ClosedFormLRMethods {
@@ -70,7 +88,7 @@ impl From<f64> for ClosedFormLRMethods {
     }
 }
 
-pub trait LinearModels {
+pub trait LinearRegression {
     fn coefficients(&self) -> MatRef<f64>;
 
     /// Returns a copy of the coefficients
@@ -80,12 +98,23 @@ pub trait LinearModels {
 
     fn bias(&self) -> f64;
 
-    fn has_bias(&self) -> bool;
-
-    fn fit(&mut self, X: MatRef<f64>, y: MatRef<f64>) -> Result<(), LinalgErrors>;
+    fn fit_bias(&self) -> bool;
 
     fn fit_unchecked(&mut self, X: MatRef<f64>, y: MatRef<f64>);
 
+    /// Fits the linear regression. Input X is any m x n matrix. Input y must be a m x 1 matrix.
+    /// Note, if there is a bias term in the data, then it must be in the matrix X as the last
+    /// column and has_bias must be true. This will not append a bias column to X.
+    fn fit(&mut self, X: MatRef<f64>, y: MatRef<f64>) -> Result<(), LinalgErrors> {
+        if X.nrows() != y.nrows() {
+            return Err(LinalgErrors::DimensionMismatch);
+        } else if X.nrows() < X.ncols() || X.nrows() == 0 || y.nrows() == 0 {
+            return Err(LinalgErrors::NotEnoughData);
+        }
+        self.fit_unchecked(X, y);
+        Ok(())
+    }
+
     fn is_fit(&self) -> bool {
         !(self.coefficients().shape() == (0, 0))
     }
@@ -116,12 +145,16 @@ pub trait LinearModels {
         } else if !self.is_fit() {
             Err(LinalgErrors::MatNotLearnedYet)
         } else {
-            if self.has_bias() && self.bias().abs() > f64::EPSILON {
-                let temp = Mat::full(X.nrows(), 1, self.bias());
-                Ok(X * self.coefficients() + temp)
-            } else {
-                Ok(X * self.coefficients())
+            let mut result = X * self.coefficients();
+            let bias = self.bias();
+            if self.fit_bias() && self.bias().abs() > f64::EPSILON {
+                unsafe {
+                    for i in 0..result.nrows() {
+                        *result.get_mut_unchecked(i, 0) += bias;
+                    }
+                }
             }
+            Ok(result)
         }
     }
 }
@@ -166,7 +199,7 @@ impl LR {
     }
 }
 
-impl LinearModels for LR {
+impl LinearRegression for LR {
     fn coefficients(&self) -> MatRef<f64> {
         self.coefficients.as_ref()
     }
@@ -175,7 +208,7 @@ impl LinearModels for LR {
         self.bias
     }
 
-    fn has_bias(&self) -> bool {
+    fn fit_bias(&self) -> bool {
         self.fit_bias
     }
 
@@ -207,20 +240,6 @@ impl LinearModels for LR {
             self.coefficients = all_coefficients;
         }
     }
-
-    /// Fits the linear regression. Input X is any m x n matrix. Input y must be a m x 1 matrix.
-    /// Note, if there is a bias term in the data, then it must be in the matrix X as the last
-    /// column and has_bias must be true. This will not append a bias column to X.
-    /// This doesn't check dimensions.
-    fn fit(&mut self, X: MatRef<f64>, y: MatRef<f64>) -> Result<(), LinalgErrors> {
-        if X.nrows() != y.nrows() {
-            return Err(LinalgErrors::DimensionMismatch);
-        } else if X.nrows() < X.ncols() {
-            return Err(LinalgErrors::NotEnoughRows);
-        }
-        self.fit_unchecked(X, y);
-        Ok(())
-    }
 }
 
 /// A struct that handles online linear regression
@@ -248,17 +267,16 @@ impl OnlineLR {
     pub fn set_coeffs_bias_inverse(
         &mut self,
         coeffs: &[f64],
-        inv: ArrayView2<f64>,
         bias: f64,
+        inv: MatRef<f64>,
     ) -> Result<(), LinalgErrors> {
         if coeffs.len() != inv.ncols() {
             Err(LinalgErrors::DimensionMismatch)
         } else {
+            self.bias = bias;
             self.coefficients =
                 (faer::mat::from_row_major_slice(coeffs, coeffs.len(), 1)).to_owned();
-            self.inv = inv.into_faer().to_owned();
-            self.bias = bias;
-            self.fit_bias = bias.abs() > f64::EPSILON;
+            self.inv = inv.to_owned();
             Ok(())
         }
     }
@@ -272,17 +290,31 @@ impl OnlineLR {
     }
 
     pub fn update_unchecked(&mut self, new_x: MatRef<f64>, new_y: MatRef<f64>, c: f64) {
-        woodbury_step(
-            self.inv.as_mut(),
-            self.coefficients.as_mut(),
-            new_x,
-            new_y,
-            c,
-        )
-    }
-
-    pub fn update(&mut self, new_x: MatRef<f64>, new_y: MatRef<f64>, c: f64) {
-        if !(new_x.has_nan() || new_y.has_nan()) {
+        if self.fit_bias() {
+            let cur_coeffs = self.coefficients();
+            let ones = Mat::full(new_x.nrows(), 1, 1.0);
+            let new_new_x = faer::concat![[new_x, ones]];
+            // Need this because of dimension issue. Coefficients doesn't contain the bias term, but
+            // during fit, it was included which resulted in +1 dimension.
+            // We need to take care of this.
+            let nfeats = cur_coeffs.nrows();
+            let mut temp_weights = Mat::<f64>::from_fn(nfeats + 1, 1, |i, j| {
+                if i < nfeats {
+                    *cur_coeffs.get(i, j)
+                } else {
+                    self.bias
+                }
+            });
+            woodbury_step(
+                self.inv.as_mut(),
+                temp_weights.as_mut(),
+                new_new_x.as_ref(),
+                new_y,
+                c,
+            );
+            self.coefficients = temp_weights.get(..nfeats, ..).to_owned();
+            self.bias = *temp_weights.get(nfeats, 0);
+        } else {
             woodbury_step(
                 self.inv.as_mut(),
                 self.coefficients.as_mut(),
@@ -292,9 +324,15 @@ impl OnlineLR {
             )
         }
     }
+
+    pub fn update(&mut self, new_x: MatRef<f64>, new_y: MatRef<f64>, c: f64) {
+        if !(new_x.has_nan() || new_y.has_nan()) {
+            self.update_unchecked(new_x, new_y, c)
+        }
+    }
 }
 
-impl LinearModels for OnlineLR {
+impl LinearRegression for OnlineLR {
     fn coefficients(&self) -> MatRef<f64> {
         self.coefficients.as_ref()
     }
@@ -303,27 +341,135 @@ impl LinearModels for OnlineLR {
         self.bias
     }
 
-    fn has_bias(&self) -> bool {
+    fn fit_bias(&self) -> bool {
         self.fit_bias
     }
 
     fn fit_unchecked(&mut self, X: MatRef<f64>, y: MatRef<f64>) {
-        (self.inv, self.coefficients) = match self.method {
-            ClosedFormLRMethods::Normal => faer_qr_lstsq_with_inv(X, y),
-            ClosedFormLRMethods::L2 => {
-                faer_cholesky_ridge_with_inv(X, y, self.lambda, self.fit_bias)
-            }
+        if self.fit_bias {
+            let actual_features = X.ncols();
+            let ones = Mat::full(X.nrows(), 1, 1.0);
+            let new_x = faer::concat![[X, ones]];
+            let (inv, all_coefficients) = match self.method {
+                ClosedFormLRMethods::Normal => faer_qr_lstsq_with_inv(new_x.as_ref(), y),
+                ClosedFormLRMethods::L2 => {
+                    faer_cholesky_ridge_with_inv(new_x.as_ref(), y, self.lambda, true)
+                }
+            };
+            self.inv = inv;
+            self.coefficients = all_coefficients.get(..actual_features, ..).to_owned();
+            self.bias = *all_coefficients.get(actual_features, 0);
+        } else {
+            (self.inv, self.coefficients) = match self.method {
+                ClosedFormLRMethods::Normal => faer_qr_lstsq_with_inv(X, y),
+                ClosedFormLRMethods::L2 => faer_cholesky_ridge_with_inv(X, y, self.lambda, false),
+            };
+        }
+    }
+}
+
+/// A struct that handles regular linear regression and Ridge regression.
+pub struct ElasticNet {
+    pub l1_reg: f64,
+    pub l2_reg: f64,
+    pub coefficients: Mat<f64>, // n_features x 1 matrix, doesn't contain bias
+    pub fit_bias: bool,
+    pub bias: f64,
+    pub tol: f64,
+    pub max_iter: usize,
+}
+
+impl ElasticNet {
+    pub fn new(l1_reg: f64, l2_reg: f64, fit_bias: bool, tol: f64, max_iter: usize) -> Self {
+        ElasticNet {
+            l1_reg: l1_reg,
+            l2_reg: l2_reg,
+            coefficients: Mat::new(),
+            fit_bias: fit_bias,
+            bias: 0.,
+            tol: tol,
+            max_iter: max_iter,
+        }
+    }
+
+    pub fn from_values(coeffs: &[f64], bias: f64) -> Self {
+        ElasticNet {
+            l1_reg: f64::NAN,
+            l2_reg: f64::NAN,
+            coefficients: faer::mat::from_row_major_slice(coeffs, coeffs.len(), 1).to_owned(),
+            fit_bias: bias.abs() > f64::EPSILON,
+            bias: bias,
+            tol: 1e-5,
+            max_iter: 2000,
+        }
+    }
+
+    pub fn set_coeffs_and_bias(&mut self, coeffs: &[f64], bias: f64) {
+        self.coefficients = (faer::mat::from_row_major_slice(coeffs, coeffs.len(), 1)).to_owned();
+        self.bias = bias;
+        self.fit_bias = bias.abs() > f64::EPSILON;
+    }
+
+    pub fn regularizers(&self) -> (f64, f64) {
+        (self.l1_reg, self.l2_reg)
+    }
+}
+
+impl LinearRegression for ElasticNet {
+    fn coefficients(&self) -> MatRef<f64> {
+        self.coefficients.as_ref()
+    }
+
+    fn bias(&self) -> f64 {
+        self.bias
+    }
+
+    fn fit_bias(&self) -> bool {
+        self.fit_bias
+    }
+
+    fn fit_unchecked(&mut self, X: MatRef<f64>, y: MatRef<f64>) {
+        let all_coefficients = if self.fit_bias {
+            let ones = Mat::full(X.nrows(), 1, 1.0);
+            let new_x = faer::concat![[X, ones]];
+            faer_coordinate_descent(
+                new_x.as_ref(),
+                y,
+                self.l1_reg,
+                self.l2_reg,
+                self.fit_bias,
+                self.tol,
+                self.max_iter,
+            )
+        } else {
+            faer_coordinate_descent(
+                X,
+                y,
+                self.l1_reg,
+                self.l2_reg,
+                self.fit_bias,
+                self.tol,
+                self.max_iter,
+            )
         };
+
+        if self.fit_bias {
+            let n = all_coefficients.nrows();
+            let slice = all_coefficients.col_as_slice(0);
+            self.coefficients =
+                faer::mat::from_row_major_slice(&slice[..n - 1], n - 1, 1).to_owned();
+            self.bias = slice[n - 1];
+        } else {
+            self.coefficients = all_coefficients;
+        }
     }
 
-    /// The very initial fit of the Online Regression Problem.
-    /// Defaults to Column Pivot QR for normal and Cholesky for Ridge
     fn fit(&mut self, X: MatRef<f64>, y: MatRef<f64>) -> Result<(), LinalgErrors> {
         if X.nrows() != y.nrows() {
             return Err(LinalgErrors::DimensionMismatch);
-        } else if X.nrows() < X.ncols() {
-            return Err(LinalgErrors::NotEnoughRows);
-        }
+        } else if X.nrows() == 0 || y.nrows() == 0 {
+            return Err(LinalgErrors::NotEnoughData);
+        } // Ok to have nrows < ncols
         self.fit_unchecked(X, y);
         Ok(())
     }
@@ -347,8 +493,7 @@ pub fn faer_solve_lstsq(x: MatRef<f64>, y: MatRef<f64>, how: LRSolverMethods) ->
 /// Least square that sets all singular values below threshold to 0.
 /// Returns the coefficients and the singular values
 #[inline(always)]
-pub fn faer_solve_lstsq_rcond(x: MatRef<f64>, y: MatRef<f64>, rcond:f64) -> (Mat<f64>, Vec<f64>) {
-
+pub fn faer_solve_lstsq_rcond(x: MatRef<f64>, y: MatRef<f64>, rcond: f64) -> (Mat<f64>, Vec<f64>) {
     let xt = x.transpose();
     let svd = (xt * x).thin_svd();
 
@@ -366,29 +511,26 @@ pub fn faer_solve_lstsq_rcond(x: MatRef<f64>, y: MatRef<f64>, rcond:f64) -> (Mat
         .s_diagonal()
         .iter()
         .copied()
-        .map(|x| if x >= threshold { x.recip() } else { 0. } )
+        .map(|x| if x >= threshold { x.recip() } else { 0. })
         .collect::<Vec<_>>();
 
     let s_inv = faer::mat::from_row_major_slice(&s_inv, s_inv.len(), 1);
     let s_inv = s_inv.column_vector_as_diagonal();
 
     let weights = svd.v() * s_inv * svd.u().transpose() * xt * y;
-    (weights, singular_values)    
-
+    (weights, singular_values)
 }
 
 /// Least square that sets all singular values below threshold to 0.
 /// Returns the coefficients and the singular values
 #[inline(always)]
 pub fn faer_solve_ridge_rcond(
-    x: MatRef<f64>, 
-    y: MatRef<f64>, 
+    x: MatRef<f64>,
+    y: MatRef<f64>,
     lambda: f64,
     has_bias: bool,
-    rcond:f64
+    rcond: f64,
 ) -> (Mat<f64>, Vec<f64>) {
-
-
     let n1 = x.ncols().abs_diff(has_bias as usize);
     let xt = x.transpose();
     let mut xtx_plus = xt * x;
@@ -413,19 +555,16 @@ pub fn faer_solve_ridge_rcond(
         .s_diagonal()
         .iter()
         .copied()
-        .map(|x| if x >= threshold { x.recip() } else { 0. } )
+        .map(|x| if x >= threshold { x.recip() } else { 0. })
         .collect::<Vec<_>>();
 
     let s_inv = faer::mat::from_row_major_slice(&s_inv, s_inv.len(), 1);
     let s_inv = s_inv.column_vector_as_diagonal();
 
     let weights = svd.v() * s_inv * svd.u().transpose() * xt * y;
-    (weights, singular_values)    
-
+    (weights, singular_values)
 }
 
-
-
 /// Returns the coefficients for lstsq with l2 (Ridge) regularization as a nrows x 1 matrix
 #[inline(always)]
 pub fn faer_solve_ridge(
@@ -469,7 +608,7 @@ pub fn faer_qr_lstsq_with_inv(x: MatRef<f64>, y: MatRef<f64>) -> (Mat<f64>, Mat<
     (inv, weights)
 }
 
-/// Returns the coefficients for lstsq with l2 (Ridge) regularization as a nrows x 1 matrix
+/// Returns the coefficients for lstsq with l2 (Ridge) regularization atogether with the inverse of XtX
 /// By default this uses Choleskey to solve the system, and in case the matrix is not positive
 /// definite, it falls back to SVD. (I suspect the matrix in this case is always positive definite!)
 #[inline(always)]
@@ -503,7 +642,12 @@ pub fn faer_cholesky_ridge_with_inv(
 
 /// Solves the weighted least square with weights given by the user
 #[inline(always)]
-pub fn faer_weighted_lstsq(x: MatRef<f64>, y: MatRef<f64>, w: &[f64], how:LRSolverMethods) -> Mat<f64> {
+pub fn faer_weighted_lstsq(
+    x: MatRef<f64>,
+    y: MatRef<f64>,
+    w: &[f64],
+    how: LRSolverMethods,
+) -> Mat<f64> {
     let weights = faer::mat::from_row_major_slice(w, x.nrows(), 1);
     let w = weights.column_vector_as_diagonal();
     let xt = x.transpose();
@@ -512,62 +656,66 @@ pub fn faer_weighted_lstsq(x: MatRef<f64>, y: MatRef<f64>, w: &[f64], how:LRSolv
         LRSolverMethods::SVD => {
             let svd = (&xtw * x).thin_svd();
             svd.solve(xtw * y)
-        },
+        }
         LRSolverMethods::QR => {
             let qr = (&xtw * x).col_piv_qr();
             qr.solve(xtw * y)
-        },
+        }
         LRSolverMethods::Choleskey => match (&xtw * x).cholesky(Side::Lower) {
             Ok(cho) => cho.solve(xtw * y),
             Err(_) => {
                 let svd = (&xtw * x).thin_svd();
                 svd.solve(xtw * y)
-            },
+            }
         },
     }
 }
 
 #[inline(always)]
-fn soft_threshold_l1(rho: f64, lambda: f64) -> f64 {
-    rho.signum() * (rho.abs() - lambda).max(0f64)
+fn soft_threshold_l1(z: f64, lambda: f64) -> f64 {
+    z.signum() * (z.abs() - lambda).max(0f64)
 }
 
-/// Computes Lasso Regression coefficients by the use of Coordinate Descent.
+/// Computes Lasso/Elastic Regression coefficients by the use of Coordinate Descent.
 /// The current stopping criterion is based on L Inf norm of the changes in the
 /// coordinates. A better alternative might be the dual gap.
 ///
 /// Reference:
 /// https://xavierbourretsicotte.github.io/lasso_implementation.html
+/// https://www.stat.cmu.edu/~ryantibs/convexopt-F18/lectures/coord-desc.pdf
 /// https://github.com/minatosato/Lasso/blob/master/coordinate_descent_lasso.py
 /// https://en.wikipedia.org/wiki/Lasso_(statistics)
 #[inline(always)]
-pub fn faer_lasso_regression(
+pub fn faer_coordinate_descent(
     x: MatRef<f64>,
     y: MatRef<f64>,
-    lambda: f64,
+    l1_reg: f64,
+    l2_reg: f64,
     has_bias: bool,
     tol: f64,
+    max_iter: usize,
 ) -> Mat<f64> {
     let m = x.nrows() as f64;
     let ncols = x.ncols();
     let n1 = ncols.abs_diff(has_bias as usize);
 
-    let lambda_new = m * lambda;
+    let lambda_l1 = m * l1_reg;
 
     let mut beta: Mat<f64> = Mat::zeros(ncols, 1);
     let mut converge = false;
 
     // compute column squared l2 norms.
+    // (In the case of Elastic net, squared l2 norms + l2 regularization factor)
     let norms = x
         .col_iter()
-        .map(|c| c.squared_norm_l2())
+        .map(|c| c.squared_norm_l2() + m * l2_reg)
         .collect::<Vec<_>>();
 
     let xty = x.transpose() * y;
     let xtx = x.transpose() * x;
 
     // Random selection often leads to faster convergence?
-    for _ in 0..2000 {
+    for _ in 0..max_iter {
         let mut max_change = 0f64;
         for j in 0..n1 {
             // temporary set beta(j, 0) to 0.
@@ -576,10 +724,11 @@ pub fn faer_lasso_regression(
             *unsafe { beta.get_mut_unchecked(j, 0) } = 0f64;
             let xtx_j = unsafe { xtx.get_unchecked(j..j + 1, ..) };
 
-            let dot = xty.read(j, 0) - (xtx_j * &beta).read(0, 0);
+            // Xi^t(y - X-i Beta-i)
+            let main_update = xty.read(j, 0) - (xtx_j * &beta).read(0, 0);
 
             // update beta(j, 0).
-            let after = soft_threshold_l1(dot, lambda_new) / norms[j];
+            let after = soft_threshold_l1(main_update, lambda_l1) / norms[j];
             *unsafe { beta.get_mut_unchecked(j, 0) } = after;
             max_change = (after - before).abs().max(max_change);
         }
@@ -598,7 +747,9 @@ pub fn faer_lasso_regression(
     }
 
     if !converge {
-        println!("Lasso regression: 2000 iterations have passed and result hasn't converged.")
+        println!(
+            "Lasso regression: Max number of iterations have passed and result hasn't converged."
+        )
     }
 
     beta
@@ -611,7 +762,6 @@ pub fn faer_recursive_lstsq(
     y: MatRef<f64>,
     n: usize,
     lambda: f64,
-    has_bias: bool,
 ) -> Vec<Mat<f64>> {
     let xn = x.nrows();
     // x: size xn x m
@@ -622,7 +772,9 @@ pub fn faer_recursive_lstsq(
     let x0 = x.get(..n, ..);
     let y0 = y.get(..n, ..);
 
-    let mut online_lr = OnlineLR::new(lambda, has_bias);
+    // This is because if add_bias, the 1 is added to
+    // all data already. No need to let OnlineLR add the 1 for the user.
+    let mut online_lr = OnlineLR::new(lambda, false);
     online_lr.fit_unchecked(x0, y0); // safe because things are checked in plugin / python functions.
     coefficients.push(online_lr.get_coefficients());
     for j in n..xn {
@@ -637,13 +789,7 @@ pub fn faer_recursive_lstsq(
 /// Given all data, we start running a lstsq starting at position n and compute new coefficients recurisively.
 /// This will return all coefficients for rows >= n. This will only be used in Polars Expressions.
 /// This supports Normal or Ridge regression
-pub fn faer_rolling_lstsq(
-    x: MatRef<f64>,
-    y: MatRef<f64>,
-    n: usize,
-    lambda: f64,
-    has_bias: bool,
-) -> Vec<Mat<f64>> {
+pub fn faer_rolling_lstsq(x: MatRef<f64>, y: MatRef<f64>, n: usize, lambda: f64) -> Vec<Mat<f64>> {
     let xn = x.nrows();
     // x: size xn x m
     // y: size xn x 1
@@ -653,7 +799,9 @@ pub fn faer_rolling_lstsq(
     let x0 = x.get(..n, ..);
     let y0 = y.get(..n, ..);
 
-    let mut online_lr = OnlineLR::new(lambda, has_bias);
+    // This is because if add_bias, the 1 is added to
+    // all data already. No need to let OnlineLR add the 1 for the user.
+    let mut online_lr = OnlineLR::new(lambda, false);
     online_lr.fit_unchecked(x0, y0);
     coefficients.push(online_lr.get_coefficients());
 
@@ -680,7 +828,6 @@ pub fn faer_rolling_skipping_lstsq(
     n: usize,
     m: usize,
     lambda: f64,
-    has_bias: bool,
 ) -> Vec<Mat<f64>> {
     let xn = x.nrows();
     let ncols = x.ncols();
@@ -697,7 +844,9 @@ pub fn faer_rolling_skipping_lstsq(
     let mut x_slice: Vec<f64> = Vec::with_capacity(n * ncols);
     let mut y_slice: Vec<f64> = Vec::with_capacity(n);
 
-    let mut online_lr = OnlineLR::new(lambda, has_bias);
+    // This is because if add_bias, the 1 is added to
+    // all data already. No need to let OnlineLR add the 1 for the user.
+    let mut online_lr = OnlineLR::new(lambda, false);
     while right <= xn {
         // Somewhat redundant here.
         non_null_cnt_in_window = 0;
@@ -762,7 +911,7 @@ fn woodbury_step(
     weights: MatMut<f64>,
     new_x: MatRef<f64>,
     new_y: MatRef<f64>,
-    c: f64, // Should be +1 or -1, for a "update" and a "removal"
+    c: f64, // +1 or -1, for a "update" and a "removal"
 ) {
     // It is truly amazing that the C in the Woodbury identity essentially controls the update and
     // and removal of a new record (rolling)... Linear regression seems to be designed by God to work so well
@@ -770,7 +919,7 @@ fn woodbury_step(
     let left = &inverse * new_x.transpose(); // corresponding to u in the reference
                                              // right = left.transpose() by the fact that if A is symmetric, invertible, A-1 is also symmetric
     let z = (c + (new_x * &left).read(0, 0)).recip();
-    // Update the inverse
+    // Update the information matrix's inverse. Page 56 of the gatech reference
     faer::linalg::matmul::matmul(
         inverse,
         &left,
@@ -782,7 +931,7 @@ fn woodbury_step(
 
     // Difference from esitmate using prior weights vs. actual next y
     let y_diff = new_y - (new_x * &weights);
-    // Update weights
+    // Update weights. Page 56, after 'Then',.. in gatech reference
     faer::linalg::matmul::matmul(
         weights,
         left,
