@@ -2,13 +2,12 @@
 /// other features/entropies that require KNN to be efficiently computed.
 use crate::{
     arkadia::{
-        matrix_to_empty_leaves, matrix_to_leaves, AnyKDT, KNNMethod, KNNRegressor, Leaf,
-        SpacialQueries,
+        utils::{slice_to_empty_leaves, slice_to_leaves},
+        KNNMethod, KNNRegressor, Leaf, SpatialQueries, KDT,
     },
-    utils::{list_u32_output, series_to_ndarray, split_offsets, DIST},
+    utils::{list_u32_output, series_to_row_major_slice, split_offsets, DIST},
 };
 
-use ndarray::{s, ArrayView2};
 use num::Float;
 use polars::prelude::*;
 use pyo3_polars::{
@@ -21,10 +20,10 @@ use pyo3_polars::{
 use serde::Deserialize;
 
 pub fn knn_full_output(_: &[Field]) -> PolarsResult<Field> {
-    let idx = Field::new("idx", DataType::List(Box::new(DataType::UInt32)));
-    let dist = Field::new("dist", DataType::List(Box::new(DataType::Float64)));
+    let idx = Field::new("idx".into(), DataType::List(Box::new(DataType::UInt32)));
+    let dist = Field::new("dist".into(), DataType::List(Box::new(DataType::Float64)));
     let v = vec![idx, dist];
-    Ok(Field::new("knn_dist", DataType::Struct(v)))
+    Ok(Field::new("knn_dist".into(), DataType::Struct(v)))
 }
 
 #[derive(Deserialize)]
@@ -69,33 +68,20 @@ pub(crate) struct KDTRadiusKwargs {
     pub(crate) sort: bool,
 }
 
-pub fn matrix_to_leaves_filtered<'a, T: Float + 'static, A: Copy>(
-    matrix: &'a ArrayView2<'a, T>,
+pub fn row_major_slice_to_leaves_filtered<'a, T: Float + 'static, A: Copy>(
+    slice: &'a [T],
+    row_len: usize,
     values: &'a [A],
     filter: &BooleanChunked,
 ) -> Vec<Leaf<'a, T, A>> {
     filter
         .into_iter()
-        .zip(values.iter().copied().zip(matrix.rows()))
+        .zip(values.iter().copied().zip(slice.chunks_exact(row_len)))
         .filter(|(f, _)| f.unwrap_or(false))
         .map(|(_, pair)| pair.into())
         .collect::<Vec<_>>()
 }
 
-// used in all cases but squared l2 (multiple queries)
-pub fn dist_from_str<T: Float + cfavml::safe_trait_distance_ops::DistanceOps + 'static>(
-    dist_str: String,
-) -> Result<DIST<T>, String> {
-    match dist_str.as_ref() {
-        "l1" => Ok(DIST::L1),
-        "l2" => Ok(DIST::L2),
-        "sql2" => Ok(DIST::SQL2),
-        "linf" | "inf" => Ok(DIST::LINF),
-        "cosine" => Ok(DIST::ANY(cfavml::cosine)),
-        _ => Err("Unknown distance metric.".into()),
-    }
-}
-
 /// KNN Regression
 /// Always do k + 1 because this operation is in-dataframe, and this means
 /// that the point itself is always a neighbor to itself.
@@ -119,44 +105,38 @@ fn pl_knn_avg(
     let null_mask = inputs[1].bool().unwrap();
     let nrows = null_mask.len();
 
-    let data = series_to_ndarray(&inputs[2..], IndexOrder::C)?;
-    let binding = data.view();
-    let mut leaves = matrix_to_leaves_filtered(&binding, id, &null_mask);
+    // let data = series_to_ndarray(&inputs[2..], IndexOrder::C)?;
 
-    let tree = match dist_from_str::<f64>(kwargs.metric) {
-        Ok(d) => Ok(AnyKDT::from_leaves_unchecked(&mut leaves, d)),
+    let ncols = inputs[2..].len();
+    let data = series_to_row_major_slice::<Float64Type>(&inputs[2..])?;
+    let mut leaves = row_major_slice_to_leaves_filtered(&data, ncols, id, &null_mask);
+
+    let tree = match DIST::<f64>::new_from_str_informed(kwargs.metric, ncols) {
+        Ok(d) => Ok(KDT::from_leaves_unchecked(&mut leaves, d)),
         Err(e) => Err(e),
     }
     .map_err(|err| PolarsError::ComputeError(err.into()))?;
 
     let ca = if can_parallel {
-        POOL.install(|| {
-            let n_threads = POOL.current_num_threads();
-            let splits = split_offsets(nrows, n_threads);
-            let chunks: Vec<_> = splits
-                .into_par_iter()
-                .map(|(offset, len)| {
-                    let slice = binding.slice(s![offset..offset + len, ..]);
-                    let out = Float64Chunked::from_iter_options(
-                        "",
-                        slice.rows().into_iter().map(|row| {
-                            let sl = row.as_slice().unwrap();
-                            tree.knn_regress(k + 1, sl, min_bound, max_bound, method)
-                        }),
-                    );
-                    out.downcast_iter().cloned().collect::<Vec<_>>()
-                })
-                .collect();
-
-            Float64Chunked::from_chunk_iter("", chunks.into_iter().flatten())
-        })
+        let n_threads = POOL.current_num_threads();
+        let splits = split_offsets(nrows, n_threads);
+        let chunks_iter = splits.into_par_iter().map(|(offset, len)| {
+            let subslice = &data[offset * ncols..(offset + len) * ncols];
+            let out = Float64Chunked::from_iter_options(
+                "".into(),
+                subslice
+                    .chunks_exact(ncols)
+                    .map(|row| tree.knn_regress(k + 1, row, min_bound, max_bound, method)),
+            );
+            out.downcast_iter().cloned().collect::<Vec<_>>()
+        });
+        let chunks = POOL.install(|| chunks_iter.collect::<Vec<_>>());
+        Float64Chunked::from_chunk_iter("".into(), chunks.into_iter().flatten())
     } else {
         Float64Chunked::from_iter_options(
-            "",
-            data.rows().into_iter().map(|row| {
-                let sl = row.as_slice().unwrap();
-                tree.knn_regress(k + 1, sl, min_bound, max_bound, method)
-            }),
+            "".into(),
+            data.chunks_exact(ncols)
+                .map(|row| tree.knn_regress(k + 1, row, min_bound, max_bound, method)),
         )
     };
 
@@ -166,34 +146,39 @@ fn pl_knn_avg(
 /// KNN Point-wise
 /// Always do k + 1 because this operation is in-dataframe, and this means
 /// that the point itself is always a neighbor to itself.
+/// Eval mask determines which values will be evaluated. Some can be skipped (Null will be returned) if user desires
 pub fn knn_ptwise<'a, Kdt>(
     tree: Kdt,
     eval_mask: Vec<bool>,
-    data: ArrayView2<'a, f64>,
+    data: &'a [f64],
     k: usize,
     can_parallel: bool,
     max_bound: f64,
     epsilon: f64,
 ) -> ListChunked
 where
-    Kdt: SpacialQueries<'a, f64, u32> + std::marker::Sync,
+    Kdt: SpatialQueries<'a, f64, u32> + std::marker::Sync,
 {
-    let nrows = data.nrows();
+    let ncols = tree.dim();
+    let nrows = data.len() / ncols;
     if can_parallel {
         let n_threads = POOL.current_num_threads();
         let splits = split_offsets(nrows, n_threads);
         let chunks_iter = splits.into_par_iter().map(|(offset, len)| {
-            let mut builder =
-                ListPrimitiveChunkedBuilder::<UInt32Type>::new("", len, k + 1, DataType::UInt32);
+            let mut builder = ListPrimitiveChunkedBuilder::<UInt32Type>::new(
+                "".into(),
+                len,
+                k + 1,
+                DataType::UInt32,
+            );
 
-            let piece = data.slice(s![offset..offset + len, ..]);
+            let subslice = &data[offset * ncols..(offset + len) * ncols];
             let mask = &eval_mask[offset..offset + len];
-            for (b, p) in mask.iter().zip(piece.rows()) {
+            for (b, p) in mask.iter().zip(subslice.chunks_exact(ncols)) {
                 if *b {
-                    match tree.knn_bounded(k + 1, p.to_slice().unwrap(), max_bound, epsilon) {
+                    match tree.knn_bounded(k + 1, p, max_bound, epsilon) {
                         Some(nbs) => {
-                            let v = nbs.into_iter().map(|nb| nb.to_item()).collect::<Vec<u32>>();
-                            builder.append_slice(&v);
+                            builder.append_iter_values(nbs.into_iter().map(|nb| nb.to_item()));
                         }
                         None => builder.append_null(),
                     }
@@ -205,21 +190,18 @@ where
             ca.downcast_iter().cloned().collect::<Vec<_>>()
         });
         let chunks = POOL.install(|| chunks_iter.collect::<Vec<_>>());
-        ListChunked::from_chunk_iter("knn", chunks.into_iter().flatten())
+        ListChunked::from_chunk_iter("knn".into(), chunks.into_iter().flatten())
     } else {
         let mut builder = ListPrimitiveChunkedBuilder::<UInt32Type>::new(
-            "knn",
+            "knn".into(),
             eval_mask.len(),
             k + 1,
             DataType::UInt32,
         );
-        for (b, p) in eval_mask.into_iter().zip(data.rows()) {
+        for (b, p) in eval_mask.into_iter().zip(data.chunks_exact(ncols)) {
             if b {
-                match tree.knn_bounded(k + 1, p.to_slice().unwrap(), max_bound, epsilon) {
-                    Some(nbs) => {
-                        let v = nbs.into_iter().map(|nb| nb.to_item()).collect::<Vec<u32>>();
-                        builder.append_slice(&v);
-                    }
+                match tree.knn_bounded(k + 1, p, max_bound, epsilon) {
+                    Some(nbs) => builder.append_iter_values(nbs.into_iter().map(|nb| nb.to_item())),
                     None => builder.append_null(),
                 }
             } else {
@@ -230,6 +212,73 @@ where
     }
 }
 
+#[polars_expr(output_type=Float64)]
+fn pl_dist_from_kth_nb(
+    inputs: &[Series],
+    context: CallerContext,
+    kwargs: KDTKwargs,
+) -> PolarsResult<Series> {
+    // Set up params
+    let k = kwargs.k;
+    let can_parallel = kwargs.parallel && !context.parallel();
+    let ncols = inputs.len();
+    let data = series_to_row_major_slice::<Float64Type>(inputs)?;
+    match DIST::<f64>::new_from_str_informed(kwargs.metric, ncols) {
+        Ok(d) => {
+            let mut leaves: Vec<Leaf<f64, ()>> = data
+                .chunks_exact(ncols)
+                .map(|sl| ((), sl).into())
+                .collect();
+            // let mut leaves = row_major_slice_to_leaves(&data, ncols, id, null_mask);
+            let tree = KDT::from_leaves_unchecked(&mut leaves, d);
+            let nrows = data.len() / ncols;
+            let ca = if can_parallel {
+                let n_threads = POOL.current_num_threads();
+                let splits = split_offsets(nrows, n_threads);
+                let chunks_iter = splits.into_par_iter()
+                    .map(|(i, n)| {
+                        let start = i * ncols;
+                        let end = (i + n) * ncols;
+                        let slice = &data[start..end];
+                        let mut builder: PrimitiveChunkedBuilder<Float64Type> 
+                            = PrimitiveChunkedBuilder::new("".into(), n);
+                        
+                        for point in slice.chunks_exact(ncols) {
+                            match tree.knn_bounded(k + 1, point, kwargs.max_bound, kwargs.epsilon) {
+                                Some(mut nbs) => {
+                                    builder.append_option(
+                                        nbs.pop().map(|nb| nb.to_dist())
+                                    );
+                                },
+                                _ => builder.append_null()
+                            }
+                        }
+                        let ca = builder.finish();
+                        ca.downcast_iter().cloned().collect::<Vec<_>>()
+                    });
+                let chunks = POOL.install(|| chunks_iter.collect::<Vec<_>>());
+                Float64Chunked::from_chunk_iter("".into(), chunks.into_iter().flatten()) 
+            } else {
+                let mut builder: PrimitiveChunkedBuilder<Float64Type> 
+                    = PrimitiveChunkedBuilder::new("".into(), nrows);
+                for point in data.chunks_exact(ncols) {
+                    match tree.knn_bounded(k + 1, point, kwargs.max_bound, kwargs.epsilon) {
+                        Some(mut nbs) => {
+                            builder.append_option(
+                                nbs.pop().map(|nb| nb.to_dist())
+                            );
+                        },
+                        _ => builder.append_null()
+                    }
+                }
+                builder.finish()
+            };
+            Ok(ca.into_series())
+        }
+        Err(e) => Err(PolarsError::ComputeError(e.into())),
+    }
+}
+
 #[polars_expr(output_type_func=list_u32_output)]
 fn pl_knn_ptwise(
     inputs: &[Series],
@@ -246,8 +295,8 @@ fn pl_knn_ptwise(
     let id = id.cont_slice()?;
     let nrows = id.len();
 
-    // True means no nulls, keep
-    let null_mask = inputs[1].bool().unwrap();
+    // True means keep
+    let keep_mask = inputs[1].bool().unwrap();
 
     let mut inputs_offset = 2;
     let eval_mask = if skip_eval {
@@ -261,43 +310,42 @@ fn pl_knn_ptwise(
         vec![true; nrows]
     };
 
-    let data = series_to_ndarray(&inputs[inputs_offset..], IndexOrder::C)?;
-    let binding = data.view();
+    let ncols = inputs[inputs_offset..].len();
+    let data = series_to_row_major_slice::<Float64Type>(&inputs[inputs_offset..])?;
 
-    let ca = match dist_from_str::<f64>(kwargs.metric) {
+    match DIST::<f64>::new_from_str_informed(kwargs.metric, ncols) {
         Ok(d) => {
-            let mut leaves = matrix_to_leaves_filtered(&binding, id, null_mask);
-            let tree = AnyKDT::from_leaves_unchecked(&mut leaves, d);
+            let mut leaves = row_major_slice_to_leaves_filtered(&data, ncols, id, keep_mask);
+            let tree = KDT::from_leaves_unchecked(&mut leaves, d);
             Ok(knn_ptwise(
                 tree,
                 eval_mask,
-                binding,
+                &data,
                 k,
                 can_parallel,
                 kwargs.max_bound,
                 kwargs.epsilon,
-            ))
+            ).into_series())
         }
-        Err(e) => Err(e),
+        Err(e) => Err(PolarsError::ComputeError(e.into())),
     }
-    .map_err(|err| PolarsError::ComputeError(err.into()))?;
 
-    Ok(ca.into_series())
 }
 
 pub fn knn_ptwise_w_dist<'a, Kdt>(
     tree: Kdt,
     eval_mask: Vec<bool>,
-    data: ArrayView2<'a, f64>,
+    data: &'a [f64],
     k: usize,
     can_parallel: bool,
     max_bound: f64,
     epsilon: f64,
 ) -> (ListChunked, ListChunked)
 where
-    Kdt: SpacialQueries<'a, f64, u32> + std::marker::Sync,
+    Kdt: SpatialQueries<'a, f64, u32> + std::marker::Sync,
 {
-    let nrows = data.nrows();
+    let ncols = tree.dim();
+    let nrows = data.len() / ncols;
     if can_parallel {
         POOL.install(|| {
             let n_threads = POOL.current_num_threads();
@@ -306,23 +354,22 @@ where
                 .into_par_iter()
                 .map(|(offset, len)| {
                     let mut builder = ListPrimitiveChunkedBuilder::<UInt32Type>::new(
-                        "",
+                        "".into(),
                         len,
                         k + 1,
                         DataType::UInt32,
                     );
                     let mut dist_builder = ListPrimitiveChunkedBuilder::<Float64Type>::new(
-                        "",
+                        "".into(),
                         len,
                         k + 1,
                         DataType::Float64,
                     );
-                    let piece = data.slice(s![offset..offset + len, ..]);
+                    let subslice = &data[offset * ncols..(offset + len) * ncols];
                     let mask = &eval_mask[offset..offset + len];
-                    for (b, p) in mask.iter().zip(piece.rows()) {
+                    for (b, p) in mask.iter().zip(subslice.chunks_exact(ncols)) {
                         if *b {
-                            match tree.knn_bounded(k + 1, p.to_slice().unwrap(), max_bound, epsilon)
-                            {
+                            match tree.knn_bounded(k + 1, p, max_bound, epsilon) {
                                 Some(nbs) => {
                                     let mut distances = Vec::with_capacity(nbs.len());
                                     let mut neighbors = Vec::with_capacity(nbs.len());
@@ -352,22 +399,27 @@ where
                 })
                 .collect();
 
-            let ca_nb = ListChunked::from_chunk_iter("idx", chunks.0.into_iter().flatten());
-            let ca_dist = ListChunked::from_chunk_iter("dist", chunks.1.into_iter().flatten());
+            let ca_nb = ListChunked::from_chunk_iter("idx".into(), chunks.0.into_iter().flatten());
+            let ca_dist =
+                ListChunked::from_chunk_iter("dist".into(), chunks.1.into_iter().flatten());
             (ca_nb, ca_dist)
         })
     } else {
-        let mut builder =
-            ListPrimitiveChunkedBuilder::<UInt32Type>::new("idx", nrows, k + 1, DataType::UInt32);
+        let mut builder = ListPrimitiveChunkedBuilder::<UInt32Type>::new(
+            "idx".into(),
+            nrows,
+            k + 1,
+            DataType::UInt32,
+        );
         let mut dist_builder = ListPrimitiveChunkedBuilder::<Float64Type>::new(
-            "dist",
+            "dist".into(),
             nrows,
             k + 1,
             DataType::Float64,
         );
-        for (b, p) in eval_mask.into_iter().zip(data.rows()) {
+        for (b, p) in eval_mask.into_iter().zip(data.chunks_exact(ncols)) {
             if b {
-                match tree.knn_bounded(k + 1, p.to_slice().unwrap(), max_bound, epsilon) {
+                match tree.knn_bounded(k + 1, p, max_bound, epsilon) {
                     Some(nbs) => {
                         // let v = nbs.into_iter().map(|nb| nb.to_item()).collect::<Vec<u32>>();
                         // builder.append_slice(&v);
@@ -424,54 +476,60 @@ fn pl_knn_ptwise_w_dist(
         vec![true; nrows]
     };
 
-    let data = series_to_ndarray(&inputs[inputs_offset..], IndexOrder::C)?;
-    let binding = data.view();
-
-    let (ca_nb, ca_dist) = match dist_from_str::<f64>(kwargs.metric) {
+    let ncols = inputs[inputs_offset..].len();
+    let data = series_to_row_major_slice::<Float64Type>(&inputs[inputs_offset..])?;
+    let (ca_nb, ca_dist) = match DIST::<f64>::new_from_str_informed(kwargs.metric, ncols) {
         Ok(d) => {
-            let mut leaves = matrix_to_leaves_filtered(&binding, id, null_mask);
-            let tree = AnyKDT::from_leaves_unchecked(&mut leaves, d);
+            let mut leaves = row_major_slice_to_leaves_filtered(&data, ncols, id, null_mask);
+            let tree = KDT::from_leaves_unchecked(&mut leaves, d);
             Ok(knn_ptwise_w_dist(
                 tree,
                 eval_mask,
-                binding,
+                &data,
                 k,
                 can_parallel,
                 kwargs.max_bound,
                 kwargs.epsilon,
             ))
         }
-        Err(e) => Err(e),
-    }
-    .map_err(|err| PolarsError::ComputeError(err.into()))?;
+        Err(e) => Err(PolarsError::ComputeError(e.into())),
+    }?;
 
-    let out = StructChunked::new("knn_dist", &[ca_nb.into_series(), ca_dist.into_series()])?;
+    let out = StructChunked::from_series(
+        "knn_dist".into(),
+        ca_nb.len(),
+        [&ca_nb.into_series(), &ca_dist.into_series()].into_iter(),
+    )?;
     Ok(out.into_series())
 }
 
 pub fn query_radius_ptwise<'a, Kdt>(
     tree: Kdt,
-    data: ArrayView2<'a, f64>,
+    data: &'a [f64],
     r: f64,
     can_parallel: bool,
     sort: bool,
 ) -> ListChunked
 where
-    Kdt: SpacialQueries<'a, f64, u32> + std::marker::Sync,
+    Kdt: SpatialQueries<'a, f64, u32> + std::marker::Sync,
 {
+    let ncols = tree.dim();
+    let nrows = data.len() / ncols;
     if can_parallel {
-        let nrows = data.nrows();
         let n_threads = POOL.current_num_threads();
         let splits = split_offsets(nrows, n_threads);
         let chunks_iter = splits.into_par_iter().map(|(offset, len)| {
-            let mut builder =
-                ListPrimitiveChunkedBuilder::<UInt32Type>::new("", len, 16, DataType::UInt32);
-            let piece = data.slice(s![offset..offset + len, ..]);
-            for p in piece.rows() {
-                let sl = p.to_slice().unwrap();
-                if let Some(v) = tree.within(sl, r, sort) {
-                    let out: Vec<u32> = v.into_iter().map(|nb| nb.to_item()).collect();
-                    builder.append_slice(&out);
+            let mut builder = ListPrimitiveChunkedBuilder::<UInt32Type>::new(
+                "".into(),
+                len,
+                16,
+                DataType::UInt32,
+            );
+
+            let subslice = &data[offset * ncols..(offset + len) * ncols];
+            for p in subslice.chunks_exact(ncols) {
+                if let Some(v) = tree.within(p, r, sort) {
+                    builder.append_iter_values(v.into_iter().map(|nb| nb.to_item()));
                 } else {
                     builder.append_null();
                 }
@@ -480,16 +538,19 @@ where
             ca.downcast_iter().cloned().collect::<Vec<_>>()
         });
         let chunks = POOL.install(|| chunks_iter.collect::<Vec<_>>());
-        ListChunked::from_chunk_iter("", chunks.into_iter().flatten())
+        ListChunked::from_chunk_iter("".into(), chunks.into_iter().flatten())
     } else {
-        let mut builder =
-            ListPrimitiveChunkedBuilder::<UInt32Type>::new("", data.len(), 16, DataType::UInt32);
+        let mut builder = ListPrimitiveChunkedBuilder::<UInt32Type>::new(
+            "".into(),
+            data.len(),
+            16,
+            DataType::UInt32,
+        );
 
-        for p in data.rows() {
-            let sl = p.to_slice().unwrap(); // C order makes sure rows are contiguous
-            if let Some(v) = tree.within(sl, r, sort) {
-                let out: Vec<u32> = v.into_iter().map(|nb| nb.to_item()).collect();
-                builder.append_slice(&out);
+        for p in data.chunks_exact(ncols) {
+            // C order (row major) makes sure rows are contiguous
+            if let Some(v) = tree.within(p, r, sort) {
+                builder.append_iter_values(v.into_iter().map(|nb| nb.to_item()));
             } else {
                 builder.append_null();
             }
@@ -512,62 +573,52 @@ fn pl_query_radius_ptwise(
     let id = inputs[0].u32()?;
     let id = id.cont_slice()?;
 
-    let data = series_to_ndarray(&inputs[1..], IndexOrder::C)?;
-    let binding = data.view();
+    let ncols = inputs[1..].len();
+    let data = series_to_row_major_slice::<Float64Type>(&inputs[1..])?;
     // Building output
-
-    let ca = match dist_from_str::<f64>(kwargs.metric) {
+    match DIST::<f64>::new_from_str_informed(kwargs.metric, ncols) {
         Ok(d) => {
-            let mut leaves = matrix_to_leaves(&binding, id);
-            let tree = AnyKDT::from_leaves_unchecked(&mut leaves, d);
-            Ok(query_radius_ptwise(
-                tree,
-                binding,
-                radius,
-                can_parallel,
-                sort,
-            ))
+            let mut leaves = slice_to_leaves(&data, ncols, id);
+            let tree = KDT::from_leaves_unchecked(&mut leaves, d);
+            Ok(query_radius_ptwise(tree, &data, radius, can_parallel, sort).into_series())
         }
-        Err(e) => Err(e),
+        Err(e) => Err(PolarsError::ComputeError(e.into())),
     }
-    .map_err(|err| PolarsError::ComputeError(err.into()))?;
-    Ok(ca.into_series())
 }
 
 #[inline]
 pub fn query_nb_cnt<'a, Kdt>(
     tree: Kdt,
-    data: ArrayView2<'a, f64>,
+    data: &'a [f64],
     r: f64,
     can_parallel: bool,
 ) -> UInt32Chunked
 where
-    Kdt: SpacialQueries<'a, f64, ()> + std::marker::Sync,
+    Kdt: SpatialQueries<'a, f64, ()> + std::marker::Sync,
 {
     // as_slice.unwrap() is safe because when we create the matrices, we specified C order.
-    let nrows = data.nrows();
-    let ncols = data.ncols();
+    let ncols = tree.dim();
+    let nrows = data.len() / ncols;
     if can_parallel {
         let splits = split_offsets(nrows, POOL.current_num_threads());
         let chunks_iter = splits.into_par_iter().map(|(offset, len)| {
             let mut builder: PrimitiveChunkedBuilder<UInt32Type> =
-                PrimitiveChunkedBuilder::new("", nrows);
+                PrimitiveChunkedBuilder::new("".into(), nrows);
 
-            let piece = data.slice(s![offset..offset + len, ..ncols]);
-            for row in piece.rows() {
-                builder.append_option(tree.within_count(row.to_slice().unwrap(), r));
+            let subslice = &data[offset * ncols..(offset + len) * ncols];
+            for row in subslice.chunks_exact(ncols) {
+                builder.append_option(tree.within_count(row, r));
             }
             let ca = builder.finish();
             ca.downcast_iter().cloned().collect::<Vec<_>>()
         });
         let chunks = POOL.install(|| chunks_iter.collect::<Vec<_>>());
-        UInt32Chunked::from_chunk_iter("cnt", chunks.into_iter().flatten())
+        UInt32Chunked::from_chunk_iter("cnt".into(), chunks.into_iter().flatten())
     } else {
         UInt32Chunked::from_iter_options(
-            "cnt",
-            data.rows()
-                .into_iter()
-                .map(|row| tree.within_count(row.to_slice().unwrap(), r)),
+            "cnt".into(),
+            data.chunks_exact(ncols)
+                .map(|row| tree.within_count(row, r)),
         )
     }
 }
@@ -575,28 +626,27 @@ where
 #[inline]
 pub fn query_nb_cnt_w_radius<'a, Kdt>(
     tree: Kdt,
-    data: ArrayView2<'a, f64>,
+    data: &'a [f64],
     radius: &Float64Chunked,
     can_parallel: bool,
 ) -> UInt32Chunked
 where
-    Kdt: SpacialQueries<'a, f64, ()> + std::marker::Sync,
+    Kdt: SpatialQueries<'a, f64, ()> + std::marker::Sync,
 {
+    let ncols = tree.dim();
+    let nrows = data.len() / ncols;
     if can_parallel {
         let radius = radius.to_vec();
-        let nrows = data.nrows();
         let n_threads = POOL.current_num_threads();
         let splits = split_offsets(nrows, n_threads);
         let chunks_iter = splits.into_par_iter().map(|(offset, len)| {
             let mut builder: PrimitiveChunkedBuilder<UInt32Type> =
-                PrimitiveChunkedBuilder::new("", nrows);
-            let piece = data.slice(s![offset..offset + len, ..]);
+                PrimitiveChunkedBuilder::new("".into(), nrows);
+            let subslice = &data[offset * ncols..(offset + len) * ncols];
             let rad = &radius[offset..offset + len];
-            for (row, r) in piece.rows().into_iter().zip(rad.iter()) {
+            for (row, r) in subslice.chunks_exact(ncols).zip(rad.iter()) {
                 match r {
-                    Some(r) => {
-                        builder.append_option(tree.within_count(row.as_slice().unwrap(), *r))
-                    }
+                    Some(r) => builder.append_option(tree.within_count(row, *r)),
                     None => builder.append_null(),
                 }
             }
@@ -604,15 +654,14 @@ where
             ca.downcast_iter().cloned().collect::<Vec<_>>()
         });
         let chunks = POOL.install(|| chunks_iter.collect::<Vec<_>>());
-        UInt32Chunked::from_chunk_iter("cnt", chunks.into_iter().flatten())
+        UInt32Chunked::from_chunk_iter("cnt".into(), chunks.into_iter().flatten())
     } else {
         UInt32Chunked::from_iter_options(
-            "cnt",
-            data.rows()
-                .into_iter()
+            "cnt".into(),
+            data.chunks_exact(ncols)
                 .zip(radius.into_iter())
                 .map(|(row, r)| match r {
-                    Some(r) => tree.within_count(row.as_slice().unwrap(), r),
+                    Some(r) => tree.within_count(row, r),
                     None => None,
                 }),
         )
@@ -626,38 +675,37 @@ fn pl_nb_cnt(inputs: &[Series], context: CallerContext, kwargs: KDTKwargs) -> Po
     let radius = inputs[0].f64()?;
     let can_parallel = kwargs.parallel && !context.parallel();
 
-    let data = series_to_ndarray(&inputs[1..], IndexOrder::C)?;
-    let nrows = data.nrows();
+    let ncols = inputs[1..].len();
+    let data = series_to_row_major_slice::<Float64Type>(&inputs[1..])?;
+    let nrows = data.len() / ncols;
 
-    let binding = data.view();
     if radius.len() == 1 {
         let r = radius.get(0).unwrap();
-        let ca = match dist_from_str::<f64>(kwargs.metric) {
+        match DIST::<f64>::new_from_str_informed(kwargs.metric, ncols) {
             Ok(d) => {
-                let mut leaves = matrix_to_empty_leaves(&binding);
-                let tree = AnyKDT::from_leaves_unchecked(&mut leaves, d);
-                Ok(query_nb_cnt(tree, data.view(), r, can_parallel))
+                let mut leaves = slice_to_empty_leaves(&data, ncols);
+                let tree = KDT::from_leaves_unchecked(&mut leaves, d);
+                Ok(
+                    query_nb_cnt(tree, &data, r, can_parallel)
+                    .with_name("cnt".into())
+                    .into_series()
+                )
             }
-            Err(e) => Err(e),
+            Err(e) => Err(PolarsError::ComputeError(e.into())),
         }
-        .map_err(|err| PolarsError::ComputeError(err.into()))?;
-        Ok(ca.with_name("cnt").into_series())
     } else if radius.len() == nrows {
-        let ca = match dist_from_str::<f64>(kwargs.metric) {
+        match DIST::<f64>::new_from_str_informed(kwargs.metric, ncols) {
             Ok(d) => {
-                let mut leaves = matrix_to_empty_leaves(&binding);
-                let tree = AnyKDT::from_leaves_unchecked(&mut leaves, d);
-                Ok(query_nb_cnt_w_radius(
-                    tree,
-                    data.view(),
-                    radius,
-                    can_parallel,
-                ))
+                let mut leaves = slice_to_empty_leaves(&data, ncols);
+                let tree = KDT::from_leaves_unchecked(&mut leaves, d);
+                Ok(
+                    query_nb_cnt_w_radius(tree, &data, radius, can_parallel)
+                    .with_name("cnt".into())
+                    .into_series()
+                )
             }
-            Err(e) => Err(e),
+            Err(e) => Err(PolarsError::ComputeError(e.into())),
         }
-        .map_err(|err| PolarsError::ComputeError(err.into()))?;
-        Ok(ca.with_name("cnt").into_series())
     } else {
         Err(PolarsError::ShapeMismatch(
             "Inputs must have the same length or one of them must be a scalar.".into(),
