@@ -1,10 +1,11 @@
-use core::f64;
-
-/// Least Squares using Faer and ndarray.
 use crate::linalg::lstsq::{
-    faer_lasso_regression, faer_recursive_lstsq, faer_rolling_lstsq, faer_rolling_skipping_lstsq, faer_solve_lstsq, faer_solve_lstsq_rcond, faer_solve_ridge, faer_solve_ridge_rcond, faer_weighted_lstsq, ClosedFormLRMethods, LRMethods
+    faer_coordinate_descent, faer_recursive_lstsq, faer_rolling_lstsq, faer_rolling_skipping_lstsq,
+    faer_solve_lstsq, faer_solve_lstsq_rcond, faer_solve_ridge, faer_solve_ridge_rcond,
+    faer_weighted_lstsq, ClosedFormLRMethods, LRMethods,
 };
 use crate::utils::{to_frame, NullPolicy};
+/// Least Squares using Faer and ndarray.
+use core::f64;
 use faer::prelude::*;
 use faer_ext::IntoFaer;
 use itertools::Itertools;
@@ -26,6 +27,15 @@ pub(crate) struct LstsqKwargs {
     pub(crate) weighted: bool,
 }
 
+#[derive(Deserialize, Debug)]
+pub(crate) struct MultiLstsqKwargs {
+    pub(crate) bias: bool,
+    pub(crate) null_policy: String,
+    pub(crate) solver: String,
+    pub(crate) last_target_idx: usize,
+    pub(crate) l2_reg: f64,
+}
+
 // Sherman-William-Woodbury (Update, online versions) LstsqKwargs
 #[derive(Deserialize, Debug)]
 pub(crate) struct SWWLstsqKwargs {
@@ -37,55 +47,48 @@ pub(crate) struct SWWLstsqKwargs {
 }
 
 fn report_output(_: &[Field]) -> PolarsResult<Field> {
-    let features = Field::new("features", DataType::String); // index of feature
-    let beta = Field::new("beta", DataType::Float64); // estimated value for this coefficient
-    let stderr = Field::new("std_err", DataType::Float64); // Std Err for this coefficient
-    let t = Field::new("t", DataType::Float64); // t value for this coefficient
-    let p = Field::new("p>|t|", DataType::Float64); // p value for this coefficient
-    let ci_lower = Field::new("0.025", DataType::Float64); // CI lower bound at 0.025
-    let ci_upper = Field::new("0.975", DataType::Float64); // CI upper bound at 0.975
+    let features = Field::new("features".into(), DataType::String); // index of feature
+    let beta = Field::new("beta".into(), DataType::Float64); // estimated value for this coefficient
+    let stderr = Field::new("std_err".into(), DataType::Float64); // Std Err for this coefficient
+    let t = Field::new("t".into(), DataType::Float64); // t value for this coefficient
+    let p = Field::new("p>|t|".into(), DataType::Float64); // p value for this coefficient
+    let ci_lower = Field::new("0.025".into(), DataType::Float64); // CI lower bound at 0.025
+    let ci_upper = Field::new("0.975".into(), DataType::Float64); // CI upper bound at 0.975
     let v: Vec<Field> = vec![features, beta, stderr, t, p, ci_lower, ci_upper]; //  ci_lower, ci_upper
-    Ok(Field::new("lstsq_report", DataType::Struct(v)))
+    Ok(Field::new("lin_reg_report".into(), DataType::Struct(v)))
 }
 
 fn pred_residue_output(_: &[Field]) -> PolarsResult<Field> {
-    let pred = Field::new("pred", DataType::Float64);
-    let residue = Field::new("resid", DataType::Float64);
+    let pred = Field::new("pred".into(), DataType::Float64);
+    let residue = Field::new("resid".into(), DataType::Float64);
     let v = vec![pred, residue];
-    Ok(Field::new("pred", DataType::Struct(v)))
+    Ok(Field::new("pred".into(), DataType::Struct(v)))
 }
 
 fn coeff_pred_output(_: &[Field]) -> PolarsResult<Field> {
-    let coeffs = Field::new("coeffs", DataType::List(Box::new(DataType::Float64)));
-    let pred = Field::new("prediction", DataType::Float64);
+    let coeffs = Field::new("coeffs".into(), DataType::List(Box::new(DataType::Float64)));
+    let pred = Field::new("prediction".into(), DataType::Float64);
     let v: Vec<Field> = vec![coeffs, pred];
-    Ok(Field::new("", DataType::Struct(v)))
+    Ok(Field::new("".into(), DataType::Struct(v)))
 }
 
 fn coeff_singular_values_output(_: &[Field]) -> PolarsResult<Field> {
-    let coeffs = Field::new("coeffs", DataType::List(Box::new(DataType::Float64)));
-    let singular_values = Field::new("singular_values", DataType::List(Box::new(DataType::Float64)));
+    let coeffs = Field::new("coeffs".into(), DataType::List(Box::new(DataType::Float64)));
+    let singular_values = Field::new(
+        "singular_values".into(),
+        DataType::List(Box::new(DataType::Float64)),
+    );
     let v: Vec<Field> = vec![coeffs, singular_values];
-    Ok(Field::new("", DataType::Struct(v)))
+    Ok(Field::new("".into(), DataType::Struct(v)))
 }
 
 fn coeff_output(_: &[Field]) -> PolarsResult<Field> {
     Ok(Field::new(
-        "coeffs",
+        "coeffs".into(),
         DataType::List(Box::new(DataType::Float64)),
     ))
 }
 
-fn infer_regression_method(l1_reg:f64, l2_reg:f64) -> LRMethods {
-    if l1_reg > 0. && l2_reg <= 0. {
-        LRMethods::L1 
-    } else if l1_reg <= 0. && l2_reg > 0. {
-        LRMethods::L2
-    } else {
-        LRMethods::Normal
-    }
-}
-
 /// Returns a Array2 ready for linear regression, and a mask, where true means the row doesn't contain null
 #[inline(always)]
 fn series_to_mat_for_lstsq(
@@ -96,12 +99,12 @@ fn series_to_mat_for_lstsq(
     let n_features = inputs.len().abs_diff(1);
 
     // minus 1 because target is also in inputs. Target is at position 0.
-    let y_has_null = inputs[0].has_validity();
-    let has_null = inputs[1..].iter().fold(false, |_, s| s.has_validity()) | y_has_null;
+    let y_has_null = inputs[0].has_nulls();
+    let has_null = inputs[1..].iter().any(|s| s.has_nulls()) | y_has_null;
 
     let mut df = to_frame(inputs)?;
     if df.is_empty() {
-        return Err(PolarsError::ComputeError("Lstsq: empty data".into()));
+        return Err(PolarsError::ComputeError("Empty data".into()));
     }
     // Add a constant column if add_bias
     if add_bias {
@@ -116,11 +119,9 @@ fn series_to_mat_for_lstsq(
             // Like ignore, skip_window takes the raw data. The actual skip is done in the underlying function in linalg.
             NullPolicy::IGNORE | NullPolicy::SKIP_WINDOW => {
                 // false, because it has nulls
-                Ok((df, BooleanChunked::from_slice("", &[false])))
+                Ok((df, BooleanChunked::from_slice("".into(), &[false])))
             }
-            NullPolicy::RAISE => Err(PolarsError::ComputeError(
-                "Lstsq: nulls found in data".into(),
-            )),
+            NullPolicy::RAISE => Err(PolarsError::ComputeError("Nulls found in data".into())),
             NullPolicy::SKIP => {
                 let init_mask = inputs[0].is_not_null(); //0 always exist
                 let mask = inputs[1..]
@@ -134,7 +135,7 @@ fn series_to_mat_for_lstsq(
                 df = df
                     .lazy()
                     .with_columns([pl::col("*")
-                        .exclude([y_name])
+                        .exclude([y_name.clone()])
                         .cast(DataType::Float64)
                         .fill_null(lit(x))])
                     .collect()?;
@@ -144,7 +145,7 @@ fn series_to_mat_for_lstsq(
                     Ok((df, init_mask))
                 } else {
                     // all filled, no nulls
-                    let mask = BooleanChunked::from_slice("", &[true]);
+                    let mask = BooleanChunked::from_slice("".into(), &[true]);
                     Ok((df, mask))
                 }
             }
@@ -152,24 +153,24 @@ fn series_to_mat_for_lstsq(
                 df = df
                     .lazy()
                     .with_columns([pl::col("*")
-                        .exclude([y_name])
+                        .exclude([y_name.clone()])
                         .cast(DataType::Float64)
                         .fill_null(lit(x))])
                     .collect()?;
 
                 if y_has_null {
                     // Unlike fill, this doesn't drop y's nulls
-                    Ok((df, BooleanChunked::from_slice("", &[false])))
+                    Ok((df, BooleanChunked::from_slice("".into(), &[false])))
                 } else {
                     // all filled, no nulls
-                    let mask = BooleanChunked::from_slice("", &[true]);
+                    let mask = BooleanChunked::from_slice("".into(), &[true]);
                     Ok((df, mask))
                 }
             }
         }
     } else {
         // In this case, the (!mask).any() is never true, which means there is no null.
-        let mask = BooleanChunked::from_slice("", &[true]);
+        let mask = BooleanChunked::from_slice("".into(), &[true]);
         Ok((df, mask))
     }?;
 
@@ -191,8 +192,8 @@ fn pl_lstsq(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series> {
 
     let solver = kwargs.solver.as_str().into();
     let weighted = kwargs.weighted;
-    let data_for_matrix = if weighted {&inputs[1..]} else {inputs};
- 
+    let data_for_matrix = if weighted { &inputs[1..] } else { inputs };
+
     match series_to_mat_for_lstsq(data_for_matrix, add_bias, null_policy) {
         Ok((mat, _)) => {
             // Solving Least Square
@@ -203,19 +204,35 @@ fn pl_lstsq(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series> {
                 let weights = weights.cont_slice().unwrap();
                 if weights.len() != mat.nrows() {
                     return Err(PolarsError::ComputeError(
-                        "Length of weights and data in X must be the same.".into(),
+                        "Shape of weights is not the same as the data.".into(),
                     ));
                 }
                 faer_weighted_lstsq(x, y, weights, solver)
             } else {
-                match infer_regression_method(kwargs.l1_reg, kwargs.l2_reg) {
+                match LRMethods::from((kwargs.l1_reg, kwargs.l2_reg)) {
                     LRMethods::Normal => faer_solve_lstsq(x, y, solver),
-                    LRMethods::L1 => faer_lasso_regression(x, y, kwargs.l1_reg, add_bias, kwargs.tol),
-                    LRMethods::L2 => faer_solve_ridge(x, y, kwargs.l2_reg, add_bias, solver)
+                    LRMethods::L1 => {
+                        faer_coordinate_descent(x, y, kwargs.l1_reg, 0., add_bias, kwargs.tol, 2000)
+                    }
+                    LRMethods::L2 => faer_solve_ridge(x, y, kwargs.l2_reg, add_bias, solver),
+                    LRMethods::ElasticNet => faer_coordinate_descent(
+                        x,
+                        y,
+                        kwargs.l1_reg,
+                        kwargs.l2_reg,
+                        add_bias,
+                        kwargs.tol,
+                        2000,
+                    ),
                 }
             };
             let mut builder: ListPrimitiveChunkedBuilder<Float64Type> =
-                ListPrimitiveChunkedBuilder::new("coeffs", 1, coeffs.nrows(), DataType::Float64);
+                ListPrimitiveChunkedBuilder::new(
+                    "coeffs".into(),
+                    1,
+                    coeffs.nrows(),
+                    DataType::Float64,
+                );
 
             builder.append_slice(coeffs.col_as_slice(0));
             let out = builder.finish();
@@ -225,21 +242,173 @@ fn pl_lstsq(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series> {
     }
 }
 
+fn series_to_mat_for_multi_lstsq(
+    inputs: &[Series],
+    last_target_idx: usize,
+    add_bias: bool,
+    null_policy: NullPolicy,
+) -> PolarsResult<Array2<f64>> {
+    let y_has_null = inputs[..last_target_idx]
+        .iter()
+        .fold(false, |acc, s| s.has_nulls() | acc);
+    let has_null = inputs[last_target_idx..].iter().any(|s| s.has_nulls()) | y_has_null;
+
+    let mut df = if has_null {
+        match null_policy {
+            NullPolicy::RAISE => Err(PolarsError::ComputeError("Nulls found in data".into())),
+
+            NullPolicy::FILL(x) => {
+                let df = DataFrame::new(inputs.iter().map(|s| s.clone().into_column()).collect())?;
+                if y_has_null {
+                    // This is because for predictions, it becomes too complicated when different targets have nulls.
+                    // There will be too many masks we need to keep track of.
+                    // This can be dealt with but I won't implement it for now.
+                    Err(PolarsError::ComputeError(
+                        "Filling null doesn't work for multi-target lstsq when there are nulls in any of the targets.".into(),
+                    ))
+                } else {
+                    let y_names = inputs[..last_target_idx]
+                        .iter()
+                        .map(|s| s.name().clone())
+                        .collect::<Vec<_>>();
+
+                    df.lazy()
+                        .with_columns([pl::col("*")
+                            .exclude(y_names)
+                            .cast(DataType::Float64)
+                            .fill_null(lit(x))])
+                        .collect()
+                }
+            }
+            _ => Err(PolarsError::ComputeError(
+                "The null policy is not supported by multi-target linear regression.".into(),
+            )),
+        }
+    } else {
+        DataFrame::new(inputs.iter().map(|s| s.clone().into_column()).collect())
+    }?;
+
+    if add_bias {
+        df = df.lazy().with_column(lit(1f64)).collect()?;
+    }
+
+    if df.is_empty() {
+        Err(PolarsError::ComputeError("Empty data".into()))
+    } else {
+        df.to_ndarray::<Float64Type>(IndexOrder::Fortran)
+    }
+}
+
+// Strictly speaking, this output type is not correct. Should be struct of last_target_idx many
+// coeff_outputs
+#[polars_expr(output_type_func=coeff_output)]
+fn pl_lstsq_multi(inputs: &[Series], kwargs: MultiLstsqKwargs) -> PolarsResult<Series> {
+    let add_bias = kwargs.bias;
+    let solver = kwargs.solver.as_str().into();
+    let last_target_idx = kwargs.last_target_idx;
+    let null_policy = NullPolicy::try_from(kwargs.null_policy)
+        .map_err(|e| PolarsError::ComputeError(e.into()))?;
+
+    let y_names = inputs[..last_target_idx]
+        .iter()
+        .map(|s| s.name())
+        .collect::<Vec<_>>();
+    let mat = series_to_mat_for_multi_lstsq(inputs, last_target_idx, add_bias, null_policy)?;
+
+    let y = mat.slice(s![.., 0..last_target_idx]).into_faer();
+    let x = mat.slice(s![.., last_target_idx..]).into_faer();
+
+    let coeffs = match LRMethods::from((0., kwargs.l2_reg)) {
+        LRMethods::Normal => Ok(faer_solve_lstsq(x, y, solver)),
+        LRMethods::L2 => Ok(faer_solve_ridge(x, y, kwargs.l2_reg, add_bias, solver)),
+        _ => Err(PolarsError::ComputeError(
+            "The method is not supported.".into(),
+        )),
+    }?;
+
+    let df_out = DataFrame::new(
+        y_names
+            .into_iter()
+            .enumerate()
+            .map(|(i, y)| {
+                let mut builder: ListPrimitiveChunkedBuilder<Float64Type> =
+                    ListPrimitiveChunkedBuilder::new(
+                        y.clone(),
+                        1,
+                        coeffs.nrows(),
+                        DataType::Float64,
+                    );
+                builder.append_slice(coeffs.col_as_slice(i));
+                let out = builder.finish();
+                out.into_column()
+            })
+            .collect::<Vec<_>>(),
+    )?;
+    
+    Ok(df_out.into_struct("coeffs".into()).into_series())
+}
+
+// Strictly speaking, this output type is not correct.
+#[polars_expr(output_type_func=pred_residue_output)]
+fn pl_lstsq_multi_pred(inputs: &[Series], kwargs: MultiLstsqKwargs) -> PolarsResult<Series> {
+    let add_bias = kwargs.bias;
+    let solver = kwargs.solver.as_str().into();
+    let last_target_idx = kwargs.last_target_idx;
+    let null_policy = NullPolicy::try_from(kwargs.null_policy)
+        .map_err(|e| PolarsError::ComputeError(e.into()))?;
+
+    let y_names = inputs[..last_target_idx]
+        .iter()
+        .map(|s| s.name())
+        .collect::<Vec<_>>();
+    let mat = series_to_mat_for_multi_lstsq(inputs, last_target_idx, add_bias, null_policy)?;
+
+    let y = mat.slice(s![.., 0..last_target_idx]).into_faer();
+    let x = mat.slice(s![.., last_target_idx..]).into_faer();
+
+    let coeffs = match LRMethods::from((0., kwargs.l2_reg)) {
+        LRMethods::Normal => Ok(faer_solve_lstsq(x, y, solver)),
+        LRMethods::L2 => Ok(faer_solve_ridge(x, y, kwargs.l2_reg, add_bias, solver)),
+        _ => Err(PolarsError::ComputeError(
+            "The method is not supported.".into(),
+        )),
+    }?;
+
+    let pred = x * &coeffs;
+    let resid = y - &pred;
+
+    let mut s = Vec::with_capacity(y_names.len() * 2);
+    for (i, y) in y_names.into_iter().enumerate() {
+        let pred_name = format!("{}_pred", y);
+        let resid_name = format!("{}_resid", y);
+        let p = Float64Chunked::from_slice(pred_name.into(), pred.col_as_slice(i));
+        let r = Float64Chunked::from_slice(resid_name.into(), resid.col_as_slice(i));
+        s.push(p.into_column());
+        s.push(r.into_column());
+    }
+    let df_out = DataFrame::new(s)?;
+    Ok(df_out.into_struct("all_preds".into()).into_series())
+}
+
 #[polars_expr(output_type_func=coeff_singular_values_output)]
 fn pl_lstsq_w_rcond(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series> {
     let add_bias = kwargs.bias;
     let null_policy = NullPolicy::try_from(kwargs.null_policy)
         .map_err(|e| PolarsError::ComputeError(e.into()))?;
 
-    let method = if kwargs.l2_reg > 0. {ClosedFormLRMethods::L2} else {ClosedFormLRMethods::Normal};
+    let method = if kwargs.l2_reg > 0. {
+        ClosedFormLRMethods::L2
+    } else {
+        ClosedFormLRMethods::Normal
+    };
 
     // Target y is at index 0
     match series_to_mat_for_lstsq(inputs, add_bias, null_policy) {
         Ok((mat, _)) => {
-            // rcond will be passed as tol 
-            let rcond = kwargs.tol.max(
-                f64::EPSILON * (inputs.len().max(mat.len())) as f64
-            );
+            // rcond will be passed as tol
+            let rcond = kwargs
+                .tol
+                .max(f64::EPSILON * (inputs.len().max(mat.len())) as f64);
 
             // Solving Least Square
             let x = mat.slice(s![.., 1..]).into_faer();
@@ -248,29 +417,44 @@ fn pl_lstsq_w_rcond(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Seri
             //     // faer_solve_ridge_rcond
             let (coeffs, singular_values) = match method {
                 ClosedFormLRMethods::Normal => faer_solve_lstsq_rcond(x, y, rcond),
-                ClosedFormLRMethods::L2 => faer_solve_ridge_rcond(x, y, kwargs.l2_reg, add_bias, rcond),
+                ClosedFormLRMethods::L2 => {
+                    faer_solve_ridge_rcond(x, y, kwargs.l2_reg, add_bias, rcond)
+                }
             };
 
             let mut builder: ListPrimitiveChunkedBuilder<Float64Type> =
-                ListPrimitiveChunkedBuilder::new("coeffs", 1, coeffs.nrows(), DataType::Float64);
+                ListPrimitiveChunkedBuilder::new(
+                    "coeffs".into(),
+                    1,
+                    coeffs.nrows(),
+                    DataType::Float64,
+                );
 
             builder.append_slice(coeffs.col_as_slice(0));
             let coeffs_ca = builder.finish();
 
             let mut sv_builder: ListPrimitiveChunkedBuilder<Float64Type> =
-                ListPrimitiveChunkedBuilder::new("singular_values", 1, singular_values.len(), DataType::Float64);
+                ListPrimitiveChunkedBuilder::new(
+                    "singular_values".into(),
+                    1,
+                    singular_values.len(),
+                    DataType::Float64,
+                );
 
             sv_builder.append_slice(&singular_values);
             let coeffs_sv = sv_builder.finish();
 
-            let ca = StructChunked::new("", &[coeffs_ca.into_series(), coeffs_sv.into_series()])?;
+            let ca = StructChunked::from_columns(
+                "".into(),
+                coeffs_sv.len(),
+                &[coeffs_ca.into_column(), coeffs_sv.into_column()],
+            )?;
             Ok(ca.into_series())
         }
         Err(e) => Err(e),
     }
 }
 
-
 #[polars_expr(output_type_func=pred_residue_output)]
 fn pl_lstsq_pred(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series> {
     let add_bias = kwargs.bias;
@@ -279,13 +463,13 @@ fn pl_lstsq_pred(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
 
     let solver = kwargs.solver.as_str().into();
     let weighted = kwargs.weighted;
-    let data_for_matrix = if weighted {&inputs[1..]} else {inputs};
+    let data_for_matrix = if weighted { &inputs[1..] } else { inputs };
 
     match series_to_mat_for_lstsq(data_for_matrix, add_bias, null_policy.clone()) {
         Ok((mat, mask)) => {
             let y = mat.slice(s![.., 0..1]).into_faer();
             let x = mat.slice(s![.., 1..]).into_faer();
-            let coeffs =if weighted {
+            let coeffs = if weighted {
                 let weights = inputs[0].f64().unwrap();
                 let weights = weights.cont_slice().unwrap();
                 if weights.len() != mat.nrows() {
@@ -295,10 +479,21 @@ fn pl_lstsq_pred(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
                 }
                 faer_weighted_lstsq(x, y, weights, solver)
             } else {
-                match infer_regression_method(kwargs.l1_reg, kwargs.l2_reg) {
+                match LRMethods::from((kwargs.l1_reg, kwargs.l2_reg)) {
                     LRMethods::Normal => faer_solve_lstsq(x, y, solver),
-                    LRMethods::L1 => faer_lasso_regression(x, y, kwargs.l1_reg, add_bias, kwargs.tol),
-                    LRMethods::L2 => faer_solve_ridge(x, y, kwargs.l2_reg, add_bias, solver)
+                    LRMethods::L1 => {
+                        faer_coordinate_descent(x, y, kwargs.l1_reg, 0., add_bias, kwargs.tol, 2000)
+                    }
+                    LRMethods::L2 => faer_solve_ridge(x, y, kwargs.l2_reg, add_bias, solver),
+                    LRMethods::ElasticNet => faer_coordinate_descent(
+                        x,
+                        y,
+                        kwargs.l1_reg,
+                        kwargs.l2_reg,
+                        add_bias,
+                        kwargs.tol,
+                        2000,
+                    ),
                 }
             };
 
@@ -309,11 +504,12 @@ fn pl_lstsq_pred(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
             // If null policy is raise and we have nulls, we won't reach here
             // If null policy is raise and we are here, then (!&mask).any() will be false.
             // No need to check null policy here.
+            // In the mask, true means is not null. In !&mask, true means is null
             let (p, r) = if (!&mask).any() {
                 let mut p_builder: PrimitiveChunkedBuilder<Float64Type> =
-                    PrimitiveChunkedBuilder::new("pred", mask.len());
+                    PrimitiveChunkedBuilder::new("pred".into(), mask.len());
                 let mut r_builder: PrimitiveChunkedBuilder<Float64Type> =
-                    PrimitiveChunkedBuilder::new("resid", mask.len());
+                    PrimitiveChunkedBuilder::new("resid".into(), mask.len());
                 let mut i: usize = 0;
                 for mm in mask.into_no_null_iter() {
                     // mask is always non-null, mm = true means it is not null
@@ -328,13 +524,13 @@ fn pl_lstsq_pred(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
                 }
                 (p_builder.finish(), r_builder.finish())
             } else {
-                let pred = Float64Chunked::from_slice("pred", pred);
-                let residue = Float64Chunked::from_slice("resid", resid);
+                let pred = Float64Chunked::from_slice("pred".into(), pred);
+                let residue = Float64Chunked::from_slice("resid".into(), resid);
                 (pred, residue)
             };
             let p = p.into_series();
             let r = r.into_series();
-            let out = StructChunked::new("", &[p, r])?;
+            let out = StructChunked::from_series("".into(), p.len(), [&p, &r].into_iter())?;
             Ok(out.into_series())
         }
         Err(e) => Err(e),
@@ -342,13 +538,13 @@ fn pl_lstsq_pred(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
 }
 
 #[polars_expr(output_type_func=report_output)]
-fn pl_lstsq_report(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series> {
+fn pl_lin_reg_report(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series> {
     let add_bias = kwargs.bias;
     let null_policy = NullPolicy::try_from(kwargs.null_policy)
         .map_err(|e| PolarsError::ComputeError(e.into()))?;
     // index 0 is target y. Skip
     let mut name_builder =
-        StringChunkedBuilder::new("features", inputs.len() + (add_bias) as usize);
+        StringChunkedBuilder::new("features".into(), inputs.len() + (add_bias) as usize);
     for s in inputs[1..].iter().map(|s| s.name()) {
         name_builder.append_value(s);
     }
@@ -412,29 +608,31 @@ fn pl_lstsq_report(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Serie
             // Finalize
             let names_ca = name_builder.finish();
             let names_series = names_ca.into_series();
-            let coeffs_series = Float64Chunked::from_slice("beta", betas);
+            let coeffs_series = Float64Chunked::from_slice("beta".into(), betas);
             let coeffs_series = coeffs_series.into_series();
-            let stderr_series = Float64Chunked::from_vec("std_err", std_err);
+            let stderr_series = Float64Chunked::from_vec("std_err".into(), std_err);
             let stderr_series = stderr_series.into_series();
-            let t_series = Float64Chunked::from_vec("t", t_values);
+            let t_series = Float64Chunked::from_vec("t".into(), t_values);
             let t_series = t_series.into_series();
-            let p_series = Float64Chunked::from_vec("p>|t|", p_values);
+            let p_series = Float64Chunked::from_vec("p>|t|".into(), p_values);
             let p_series = p_series.into_series();
-            let lower = Float64Chunked::from_vec("0.025", ci_lower);
+            let lower = Float64Chunked::from_vec("0.025".into(), ci_lower);
             let lower = lower.into_series();
-            let upper = Float64Chunked::from_vec("0.975", ci_upper);
+            let upper = Float64Chunked::from_vec("0.975".into(), ci_upper);
             let upper = upper.into_series();
-            let out = StructChunked::new(
-                "lstsq_report",
-                &[
-                    names_series,
-                    coeffs_series,
-                    stderr_series,
-                    t_series,
-                    p_series,
-                    lower,
-                    upper,
-                ],
+            let out = StructChunked::from_series(
+                "lin_reg_report".into(),
+                names_series.len(),
+                [
+                    &names_series,
+                    &coeffs_series,
+                    &stderr_series,
+                    &t_series,
+                    &p_series,
+                    &lower,
+                    &upper,
+                ]
+                .into_iter(),
             )?;
             Ok(out.into_series())
         }
@@ -452,7 +650,7 @@ fn pl_wls_report(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
     let weights = weights.cont_slice().unwrap();
     // index 0 is weights, 1 is target y. Skip them
     let mut name_builder =
-        StringChunkedBuilder::new("features", inputs.len() + (add_bias) as usize);
+        StringChunkedBuilder::new("features".into(), inputs.len() + (add_bias) as usize);
     for s in inputs[2..].iter().map(|s| s.name()) {
         name_builder.append_value(s);
     }
@@ -521,29 +719,31 @@ fn pl_wls_report(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
             // Finalize
             let names_ca = name_builder.finish();
             let names_series = names_ca.into_series();
-            let coeffs_series = Float64Chunked::from_slice("beta", betas);
+            let coeffs_series = Float64Chunked::from_slice("beta".into(), betas);
             let coeffs_series = coeffs_series.into_series();
-            let stderr_series = Float64Chunked::from_vec("std_err", std_err);
+            let stderr_series = Float64Chunked::from_vec("std_err".into(), std_err);
             let stderr_series = stderr_series.into_series();
-            let t_series = Float64Chunked::from_vec("t", t_values);
+            let t_series = Float64Chunked::from_vec("t".into(), t_values);
             let t_series = t_series.into_series();
-            let p_series = Float64Chunked::from_vec("p>|t|", p_values);
+            let p_series = Float64Chunked::from_vec("p>|t|".into(), p_values);
             let p_series = p_series.into_series();
-            let lower = Float64Chunked::from_vec("0.025", ci_lower);
+            let lower = Float64Chunked::from_vec("0.025".into(), ci_lower);
             let lower = lower.into_series();
-            let upper = Float64Chunked::from_vec("0.975", ci_upper);
+            let upper = Float64Chunked::from_vec("0.975".into(), ci_upper);
             let upper = upper.into_series();
-            let out = StructChunked::new(
-                "lstsq_report",
-                &[
-                    names_series,
-                    coeffs_series,
-                    stderr_series,
-                    t_series,
-                    p_series,
-                    lower,
-                    upper,
-                ],
+            let out = StructChunked::from_series(
+                "lin_reg_report".into(),
+                names_series.len(),
+                [
+                    &names_series,
+                    &coeffs_series,
+                    &stderr_series,
+                    &t_series,
+                    &p_series,
+                    &lower,
+                    &upper,
+                ]
+                .into_iter(),
             )?;
             Ok(out.into_series())
         }
@@ -556,36 +756,33 @@ fn pl_wls_report(inputs: &[Series], kwargs: LstsqKwargs) -> PolarsResult<Series>
 #[polars_expr(output_type_func=coeff_pred_output)]
 fn pl_recursive_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult<Series> {
     let n = kwargs.n; // Gauranteed n >= 1
-    let has_bias = kwargs.bias;
-    
+    let add_bias = kwargs.bias;
+
     // Gauranteed in Python that this won't be SKIP. SKIP doesn't work now.
     let null_policy = NullPolicy::try_from(kwargs.null_policy)
         .map_err(|e| PolarsError::ComputeError(e.into()))?;
 
     // Target y is at index 0
-    match series_to_mat_for_lstsq(inputs, has_bias, null_policy) {
+    match series_to_mat_for_lstsq(inputs, add_bias, null_policy) {
         Ok((mat, mask)) => {
             // Solving Least Square
             let x = mat.slice(s![.., 1..]).into_faer();
             let y = mat.slice(s![.., 0..1]).into_faer();
 
-            let coeffs = faer_recursive_lstsq(x, y, n, kwargs.lambda, has_bias);
+            let coeffs = faer_recursive_lstsq(x, y, n, kwargs.lambda);
             let mut builder: ListPrimitiveChunkedBuilder<Float64Type> =
                 ListPrimitiveChunkedBuilder::new(
-                    "coeffs",
+                    "coeffs".into(),
                     mat.nrows(),
                     mat.ncols(),
                     DataType::Float64,
                 );
             let mut pred_builder: PrimitiveChunkedBuilder<Float64Type> =
-                PrimitiveChunkedBuilder::new("pred", mat.nrows());
-
-            // Note, if has_null is true, and null policy is raise, then we won't reach here.
-            // So in the has_null branch, only possibility is that null_policy != Raise
+                PrimitiveChunkedBuilder::new("pred".into(), mat.nrows());
 
             // Fill or Skip strategy can drop nulls. Fill will drop null when y has nulls.
             // Skip will drop nulls whenever there is a null in the row.
-            // Mask true means the row is good
+            // Mask true means the row doesn't have nulls
             let has_nulls = (!&mask).any();
             if has_nulls {
                 // Find the first index where we get n non-nulls.
@@ -635,7 +832,11 @@ fn pl_recursive_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult
 
             let coef_out = builder.finish();
             let pred_out = pred_builder.finish();
-            let ca = StructChunked::new("", &[coef_out.into_series(), pred_out.into_series()])?;
+            let ca = StructChunked::from_series(
+                "".into(),
+                coef_out.len(),
+                [&coef_out.into_series(), &pred_out.into_series()].into_iter(),
+            )?;
             Ok(ca.into_series())
         }
         Err(e) => Err(e),
@@ -645,7 +846,7 @@ fn pl_recursive_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult
 #[polars_expr(output_type_func=coeff_pred_output)] // They share the same output type
 fn pl_rolling_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult<Series> {
     let n = kwargs.n; // Gauranteed n >= 2
-    let has_bias = kwargs.bias;
+    let add_bias = kwargs.bias;
 
     let mut null_policy = NullPolicy::try_from(kwargs.null_policy)
         .map_err(|e| PolarsError::ComputeError(e.into()))?;
@@ -658,7 +859,7 @@ fn pl_rolling_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult<S
     };
 
     // Target y is at index 0
-    match series_to_mat_for_lstsq(inputs, has_bias, null_policy) {
+    match series_to_mat_for_lstsq(inputs, add_bias, null_policy) {
         Ok((mat, mask)) => {
             let should_skip = match null_policy {
                 NullPolicy::SKIP_WINDOW | NullPolicy::FILL_WINDOW(_) => (!&mask).any(),
@@ -666,20 +867,21 @@ fn pl_rolling_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult<S
             };
             let x = mat.slice(s![.., 1..]).into_faer();
             let y = mat.slice(s![.., 0..1]).into_faer();
-            let coeffs = match should_skip {
-                true => faer_rolling_skipping_lstsq(x, y, n, kwargs.min_size, kwargs.lambda, has_bias),
-                false => faer_rolling_lstsq(x, y, n, kwargs.lambda, has_bias),
+            let coeffs = if should_skip {
+                faer_rolling_skipping_lstsq(x, y, n, kwargs.min_size, kwargs.lambda)
+            } else {
+                faer_rolling_lstsq(x, y, n, kwargs.lambda)
             };
 
             let mut builder: ListPrimitiveChunkedBuilder<Float64Type> =
                 ListPrimitiveChunkedBuilder::new(
-                    "coeffs",
+                    "coeffs".into(),
                     mat.nrows(),
                     mat.ncols(),
                     DataType::Float64,
                 );
             let mut pred_builder: PrimitiveChunkedBuilder<Float64Type> =
-                PrimitiveChunkedBuilder::new("pred", mat.nrows());
+                PrimitiveChunkedBuilder::new("pred".into(), mat.nrows());
 
             let m = n - 1; // n >= 2 guaranteed in Python
             for _ in 0..m {
@@ -687,7 +889,6 @@ fn pl_rolling_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult<S
                 pred_builder.append_null();
             }
 
-            // Strictly speaking I don't need this branch
             if should_skip {
                 // Skipped rows will have coeffs with shape (0, 0)
                 for (i, coefficients) in coeffs.into_iter().enumerate() {
@@ -715,7 +916,11 @@ fn pl_rolling_lstsq(inputs: &[Series], kwargs: SWWLstsqKwargs) -> PolarsResult<S
 
             let coef_out = builder.finish();
             let pred_out = pred_builder.finish();
-            let ca = StructChunked::new("", &[coef_out.into_series(), pred_out.into_series()])?;
+            let ca = StructChunked::from_series(
+                "".into(),
+                coef_out.len(),
+                [&coef_out.into_series(), &pred_out.into_series()].into_iter(),
+            )?;
             Ok(ca.into_series())
         }
         Err(e) => Err(e),
