@@ -1,11 +1,13 @@
 mod consts;
 mod fuzz;
+mod generic_str_distancer;
 mod hamming;
 mod inflections;
 mod is_stopword;
 mod jaro;
-mod knn_strs;
+mod lcs_seq;
 mod levenshtein;
+mod nearest_str;
 mod osa;
 mod overlap;
 mod snowball;
