@@ -6,6 +6,25 @@ import numpy as np
 import polars_ds as pds
 from polars.testing import assert_frame_equal, assert_series_equal
 
+def test_mcc():
+    from sklearn.metrics import matthews_corrcoef
+
+    df = (
+        pds.frame(size=2000)
+        .select(
+            pds.random(0.0, 1.0).alias("x1"),
+            pds.random(0.0, 1.0).alias("x2"),
+        )
+        .with_columns(
+            pl.col("x1").round(),
+            pl.col("x2").round(),
+        )
+    )
+
+    sklearn_result = matthews_corrcoef(df["x1"], df["x2"])
+    result = df.select(pds.query_mcc("x1", "x2")).item(0, 0)
+    assert np.isclose(result, sklearn_result)
+
 
 def test_pca():
     from sklearn.decomposition import PCA
@@ -16,7 +35,7 @@ def test_pca():
         pds.random(0.0, 1.0).alias("x3"),
     )
 
-    singular_values = df.select(pds.query_singular_values("x1", "x2", "x3").alias("res"))["res"][
+    singular_values = df.select(pds.singular_values("x1", "x2", "x3").alias("res"))["res"][
         0
     ].to_numpy()
 
@@ -27,11 +46,11 @@ def test_pca():
 
     assert np.isclose(singular_values, ans_singular_values).all()
 
-    singular_values = df.select(pds.query_singular_values("x1", "x2", "x3").alias("res"))["res"][
+    singular_values = df.select(pds.singular_values("x1", "x2", "x3").alias("res"))["res"][
         0
     ].to_numpy()
 
-    vectors = df.select(pds.query_pca("x1", "x2", "x3").alias("vectors")).unnest("vectors")[
+    vectors = df.select(pds.pca("x1", "x2", "x3").alias("vectors")).unnest("vectors")[
         "weight_vector"
     ]
 
@@ -157,7 +176,7 @@ def test_longest_streak_2(a, value, res):
 
 
 @pytest.mark.parametrize(
-    "df, ft, res_full, res_valid, res_same",
+    "df, kernel, res_full, res_valid, res_same",
     [
         (
             pl.DataFrame({"a": [5, 6, 7, 8, 9]}),
@@ -168,40 +187,40 @@ def test_longest_streak_2(a, value, res):
         ),
     ],
 )
-def test_convolve(df, ft, res_full, res_valid, res_same):
-    res = df.select(pds.convolve("a", ft, mode="full"))
+def test_convolve(df, kernel, res_full, res_valid, res_same):
+    res = df.select(pds.convolve("a", kernel, mode="full"))
 
     assert_frame_equal(res, res_full)
 
-    res = df.select(pds.convolve("a", ft, mode="valid"))
+    res = df.select(pds.convolve("a", kernel, mode="valid"))
 
     assert_frame_equal(res, res_valid)
 
-    res = df.select(pds.convolve("a", ft, mode="same"))
+    res = df.select(pds.convolve("a", kernel, mode="same"))
 
     assert_frame_equal(res, res_same)
 
-    res = df.select(pds.convolve("a", ft, mode="full", parallel=True))
+    res = df.select(pds.convolve("a", kernel, mode="full", parallel=True))
 
     assert_frame_equal(res, res_full)
 
-    res = df.select(pds.convolve("a", ft, mode="valid", parallel=True))
+    res = df.select(pds.convolve("a", kernel, mode="valid", parallel=True))
 
     assert_frame_equal(res, res_valid)
 
-    res = df.select(pds.convolve("a", ft, mode="same", parallel=True))
+    res = df.select(pds.convolve("a", kernel, mode="same", parallel=True))
 
     assert_frame_equal(res, res_same)
 
-    res = df.select(pds.convolve("a", ft, mode="full", method="fft"))
+    res = df.select(pds.convolve("a", kernel, mode="full", method="fft"))
 
     assert_frame_equal(res, res_full)
 
-    res = df.select(pds.convolve("a", ft, mode="valid", method="fft"))
+    res = df.select(pds.convolve("a", kernel, mode="valid", method="fft"))
 
     assert_frame_equal(res, res_valid)
 
-    res = df.select(pds.convolve("a", ft, mode="same", method="fft"))
+    res = df.select(pds.convolve("a", kernel, mode="same", method="fft"))
 
     assert_frame_equal(res, res_same)
 
@@ -259,7 +278,7 @@ def test_fft(arr, n):
 def test_f_test(df):
     from sklearn.feature_selection import f_classif
 
-    res = df.select(pds.query_f_test(pl.col("a"), group=pl.col("target")))
+    res = df.select(pds.f_test(pl.col("a"), group=pl.col("target")))
     res = res.item(0, 0)  # A dictionary
     statistic = res["statistic"]
     pvalue = res["pvalue"]
@@ -288,7 +307,7 @@ def test_f_test(df):
 def test_mann_whitney_u(df):
     from scipy.stats import mannwhitneyu
 
-    res = df.select(pds.query_mann_whitney_u("x1", "x2"))
+    res = df.select(pds.mann_whitney_u("x1", "x2"))
     res = res.item(0, 0)  # A dictionary
     res_statistic = res["statistic"]
     res_pvalue = res["pvalue"]
@@ -365,9 +384,9 @@ def test_fract(df, res):
     ],
 )
 def test_gcd(df, other, res):
-    assert_frame_equal(df.select(pds.query_gcd("a", other).cast(pl.Int64)), res)
+    assert_frame_equal(df.select(pds.gcd("a", other).cast(pl.Int64)), res)
 
-    assert_frame_equal(df.lazy().select(pds.query_gcd("a", other).cast(pl.Int64)).collect(), res)
+    assert_frame_equal(df.lazy().select(pds.gcd("a", other).cast(pl.Int64)).collect(), res)
 
 
 @pytest.mark.parametrize(
@@ -386,9 +405,9 @@ def test_gcd(df, other, res):
     ],
 )
 def test_lcm(df, other, res):
-    assert_frame_equal(df.select(pds.query_lcm("a", other).cast(pl.Int64)), res)
+    assert_frame_equal(df.select(pds.lcm("a", other).cast(pl.Int64)), res)
 
-    assert_frame_equal(df.lazy().select(pds.query_lcm("a", other).cast(pl.Int64)).collect(), res)
+    assert_frame_equal(df.lazy().select(pds.lcm("a", other).cast(pl.Int64)).collect(), res)
 
 
 @pytest.mark.parametrize(
@@ -469,32 +488,12 @@ def test_cross_entropy(df, res):
     ],
 )
 def test_jaccard_row(df, res):
-    assert_frame_equal(df.select(pds.query_jaccard_row("a", "b").alias("res")), res)
+    assert_frame_equal(df.select(pds.jaccard_row("a", "b").alias("res")), res)
 
-    assert_frame_equal(
-        df.lazy().select(pds.query_jaccard_row("a", "b").alias("res")).collect(), res
-    )
+    assert_frame_equal(df.lazy().select(pds.jaccard_row("a", "b").alias("res")).collect(), res)
 
 
-# Hard to write generic tests because ncols can vary in X
-def test_lstsq():
-    df = pl.DataFrame({"y": [1, 2, 3, 4, 5], "a": [2, 3, 4, 5, 6], "b": [-1, -1, -1, -1, -1]})
-    res = pl.DataFrame({"y": [[1.0, 1.0]]})
-    assert_frame_equal(
-        df.select(pds.query_lstsq(pl.col("a"), pl.col("b"), target="y", add_bias=False)), res
-    )
-
-    df = pl.DataFrame(
-        {
-            "y": [1, 2, 3, 4, 5],
-            "a": [2, 3, 4, 5, 6],
-        }
-    )
-    res = pl.DataFrame({"y": [[1.0, -1.0]]})
-    assert_frame_equal(df.select(pds.query_lstsq(pl.col("a"), target="y", add_bias=True)), res)
-
-
-def test_lstsq_against_sklearn():
+def test_lin_reg_against_sklearn():
     # Random data + noise
     df = (
         pds.frame(size=5_000)
@@ -520,7 +519,7 @@ def test_lstsq_against_sklearn():
 
     # pds, normal, with bias
     normal_coeffs = df.select(
-        pds.query_lstsq(
+        pds.lin_reg(
             "x1",
             "x2",
             "x3",
@@ -539,7 +538,7 @@ def test_lstsq_against_sklearn():
 
     # pds, normal, with bias
     normal_coeffs = df.select(
-        pds.query_lstsq(
+        pds.lin_reg(
             "x1",
             "x2",
             "x3",
@@ -578,37 +577,76 @@ def test_lasso_regression():
 
     for lambda_ in [0.01, 0.05, 0.1, 0.2]:
         df_res = df.select(
-            pds.query_lstsq("x1", "x2", "x3", target="y", l1_reg=lambda_, add_bias=False).alias(
+            pds.lin_reg("x1", "x2", "x3", target="y", l1_reg=lambda_, add_bias=False).alias(
                 "coeffs"
             )
         ).explode("coeffs")
 
         res = df_res["coeffs"].to_numpy()
 
-        reg = linear_model.Lasso(alpha=lambda_, fit_intercept=False)
-        reg.fit(x, y)
-        res_sklearn = np.asarray(reg.coef_)
+        sklearn = linear_model.Lasso(alpha=lambda_, fit_intercept=False)
+        sklearn.fit(x, y)
+        res_sklearn = np.asarray(sklearn.coef_)
         assert np.all(np.abs(res_sklearn - res) < 1e-4)
 
     for lambda_ in [0.01, 0.05, 0.1, 0.2]:
         df_res = df.select(
-            pds.query_lstsq("x1", "x2", "x3", target="y", l1_reg=lambda_, add_bias=True).alias(
-                "coeffs"
-            )
+            pds.lin_reg("x1", "x2", "x3", target="y", l1_reg=lambda_, add_bias=True).alias("coeffs")
         ).explode("coeffs")
 
         res = df_res["coeffs"].to_numpy()
         res_coef = res[:3]
         res_bias = res[-1]
 
-        reg = linear_model.Lasso(alpha=lambda_, fit_intercept=True)
-        reg.fit(x, y)
-        res_sklearn = np.asarray(reg.coef_)
+        sklearn = linear_model.Lasso(alpha=lambda_, fit_intercept=True)
+        sklearn.fit(x, y)
+        res_sklearn = np.asarray(sklearn.coef_)
         assert np.all(np.abs(res_sklearn - res_coef) < 1e-4)
-        assert abs(res_bias - reg.intercept_) < 1e-4
+        assert abs(res_bias - sklearn.intercept_) < 1e-4
 
 
-def test_recursive_lstsq():
+def test_elastic_net_regression():
+    # These tests have bigger precision tolerance because of different stopping criterions
+
+    from sklearn import linear_model
+
+    df = (
+        pds.frame(size=5_000)
+        .select(
+            pds.random(0.0, 1.0).alias("x1"),
+            pds.random(0.0, 1.0).alias("x2"),
+            pds.random(0.0, 1.0).alias("x3"),
+        )
+        .with_columns(
+            y=pl.col("x1") * 0.5 + pl.col("x2") * 0.25 - pl.col("x3") * 0.15 + pds.random() * 0.0001
+        )
+    )
+
+    x = df.select("x1", "x2", "x3").to_numpy()
+    y = df["y"].to_numpy()
+
+    for reg in [0.01, 0.05, 0.1, 0.2]:
+        l1_reg = reg
+        l2_reg = reg
+
+        df_res = df.select(
+            pds.lin_reg(
+                "x1", "x2", "x3", target="y", l1_reg=l1_reg, l2_reg=l2_reg, add_bias=False
+            ).alias("coeffs")
+        ).explode("coeffs")
+
+        res = df_res["coeffs"].to_numpy()
+
+        alpha = l1_reg + l2_reg
+        l1_ratio = l1_reg / (l1_reg + l2_reg)
+
+        sklearn = linear_model.ElasticNet(alpha=alpha, l1_ratio=l1_ratio, fit_intercept=False)
+        sklearn.fit(x, y)
+        res_sklearn = np.asarray(sklearn.coef_)
+        assert np.all(np.abs(res_sklearn - res) < 1e-4)
+
+
+def test_recursive_lin_reg():
     # Test against the lstsq method with a fit whenver a new row is in the data
     size = 1_000
     df = (
@@ -630,7 +668,7 @@ def test_recursive_lstsq():
 
     df_recursive_lr = df.select(
         "y",
-        pds.query_recursive_lstsq(
+        pds.recursive_lin_reg(
             "x1",
             "x2",
             "x3",
@@ -641,7 +679,7 @@ def test_recursive_lstsq():
 
     for i in range(start_with, 30):
         coefficients = df.limit(i).select(
-            pds.query_lstsq(
+            pds.lin_reg(
                 "x1",
                 "x2",
                 "x3",
@@ -677,7 +715,7 @@ def test_recursive_ridge():
 
     df_recursive_lr = df.select(
         "y",
-        pds.query_recursive_lstsq(
+        pds.recursive_lin_reg(
             "x1",
             "x2",
             "x3",
@@ -689,7 +727,7 @@ def test_recursive_ridge():
 
     for i in range(start_with, 30):
         coefficients = df.limit(i).select(
-            pds.query_lstsq(
+            pds.lin_reg(
                 "x1",
                 "x2",
                 "x3",
@@ -704,7 +742,7 @@ def test_recursive_ridge():
         assert np.all(np.abs(normal_result - recursive_result) < 1e-5)
 
 
-def test_rolling_lstsq():
+def test_rolling_lin_reg():
     # Test rolling lstsq by comparing it with a manually rolled lstsq result.
     # Test on multiple window sizes
     size = 500
@@ -728,7 +766,7 @@ def test_rolling_lstsq():
         df_to_test = df.select(
             "id",
             "y",
-            pds.query_rolling_lstsq(
+            pds.rolling_lin_reg(
                 "x1",
                 "x2",
                 "x3",
@@ -741,9 +779,7 @@ def test_rolling_lstsq():
         results = []
         for i in range(len(df) - window_size + 1):
             temp = df.slice(i, length=window_size)
-            results.append(
-                temp.select(pds.query_lstsq("x1", "x2", "x3", target="y").alias("coeffs"))
-            )
+            results.append(temp.select(pds.lin_reg("x1", "x2", "x3", target="y").alias("coeffs")))
 
         df_answer = pl.concat(results)
         assert_frame_equal(df_to_test, df_answer)
@@ -777,7 +813,7 @@ def test_rolling_null_skips():
     min_valid_rows = 5
 
     result = df.with_columns(
-        pds.query_rolling_lstsq(
+        pds.rolling_lin_reg(
             "x1",
             "x2",
             "x3",
@@ -804,7 +840,7 @@ def test_rolling_null_skips():
 
 
 def test_rolling_ridge():
-    # Test rolling lstsq by comparing it with a manually rolled lstsq result.
+    # Test rolling linear regression by comparing it with a manually rolled result.
     # Test on multiple window sizes
     size = 500
     df = (
@@ -827,7 +863,7 @@ def test_rolling_ridge():
         df_to_test = df.select(
             "id",
             "y",
-            pds.query_rolling_lstsq(
+            pds.rolling_lin_reg(
                 "x1",
                 "x2",
                 "x3",
@@ -842,16 +878,14 @@ def test_rolling_ridge():
         for i in range(len(df) - window_size + 1):
             temp = df.slice(i, length=window_size)
             results.append(
-                temp.select(
-                    pds.query_lstsq("x1", "x2", "x3", l2_reg=0.1, target="y").alias("coeffs")
-                )
+                temp.select(pds.lin_reg("x1", "x2", "x3", l2_reg=0.1, target="y").alias("coeffs"))
             )
 
         df_answer = pl.concat(results)
         assert_frame_equal(df_to_test, df_answer)
 
 
-def test_lstsq_skip_null():
+def test_lin_reg_skip_null():
     df = pl.DataFrame(
         {
             "y": [None, 9.5, 10.5, 11.5, 12.5],
@@ -867,7 +901,7 @@ def test_lstsq_skip_null():
     )
     assert_frame_equal(
         df.select(
-            pds.query_lstsq(
+            pds.lin_reg(
                 pl.col("a"), pl.col("b"), target="y", return_pred=True, null_policy="skip"
             ).alias("result")
         ).unnest("result"),
@@ -875,7 +909,7 @@ def test_lstsq_skip_null():
     )
 
 
-def test_lstsq_in_group_by():
+def test_lin_reg_in_group_by():
     df = pl.DataFrame(
         {
             "A": [1] * 4 + [2] * 4,
@@ -886,13 +920,13 @@ def test_lstsq_in_group_by():
     )
 
     first = df.filter(pl.col("A").eq(1)).with_columns(
-        pds.query_lstsq(
+        pds.lin_reg(
             pl.col("X1"), pl.col("X2"), target=pl.col("Y"), add_bias=False, return_pred=True
         ).alias("pred")
     )
 
     second = df.filter(pl.col("A").eq(2)).with_columns(
-        pds.query_lstsq(
+        pds.lin_reg(
             pl.col("X1"), pl.col("X2"), target=pl.col("Y"), add_bias=False, return_pred=True
         ).alias("pred")
     )
@@ -903,7 +937,7 @@ def test_lstsq_in_group_by():
             "Y",
             "X1",
             "X2",
-            pds.query_lstsq(
+            pds.lin_reg(
                 pl.col("X1"), pl.col("X2"), target=pl.col("Y"), add_bias=False, return_pred=True
             ).alias("pred"),
         )
@@ -917,7 +951,7 @@ def test_lstsq_in_group_by():
     assert_frame_equal(second, test_second)
 
 
-def test_lstsq_with_rcond():
+def test_lin_reg_with_rcond():
     import numpy as np
 
     size = 5000
@@ -940,7 +974,7 @@ def test_lstsq_with_rcond():
     np_coeffs = np_coeffs.flatten()
 
     res = df.select(
-        pds.query_lstsq_w_rcond(
+        pds.lin_reg_w_rcond(
             "x1",
             "x2",
             "x3",
@@ -965,10 +999,10 @@ def test_lstsq_with_rcond():
     ],
 )
 def test_jaccard_col(df, res):
-    assert_frame_equal(df.select(pds.query_jaccard_col("a", "b").alias("j")), res)
+    assert_frame_equal(df.select(pds.jaccard_col("a", "b").alias("j")), res)
 
     assert_frame_equal(
-        df.lazy().select(df.select(pds.query_jaccard_col("a", "b").alias("j"))).collect(), res
+        df.lazy().select(df.select(pds.jaccard_col("a", "b").alias("j"))).collect(), res
     )
 
 
@@ -1110,7 +1144,7 @@ def test_ks_stats():
 
     stats = ks_2samp(a, b).statistic
     # Only statistic for now
-    res = df.select(pds.query_ks_2samp("a", "b").struct.field("statistic")).item(0, 0)
+    res = df.select(pds.ks_2samp("a", "b").struct.field("statistic")).item(0, 0)
 
     assert np.isclose(stats, res)
 
@@ -1152,7 +1186,7 @@ def test_knn_entropy(df, k, dist, res):
 def test_ttest_ind(df, eq_var):
     from scipy.stats import ttest_ind
 
-    res = df.select(pds.query_ttest_ind("a", "b", equal_var=eq_var))
+    res = df.select(pds.ttest_ind("a", "b", equal_var=eq_var))
     res = res.item(0, 0)  # A dictionary
     statistic = res["statistic"]
     pvalue = res["pvalue"]
@@ -1179,7 +1213,7 @@ def test_ttest_ind(df, eq_var):
 def test_welch_t(df):
     from scipy.stats import ttest_ind
 
-    res = df.select(pds.query_ttest_ind("a", "b", equal_var=False))
+    res = df.select(pds.ttest_ind("a", "b", equal_var=False))
     res = res.item(0, 0)  # A dictionary
     statistic = res["statistic"]
     pvalue = res["pvalue"]
@@ -1209,7 +1243,7 @@ def test_chi2(df):
     import pandas as pd
     from scipy.stats import chi2_contingency
 
-    res = df.select(pds.query_chi2("x", "y")).item(0, 0)
+    res = df.select(pds.chi2("x", "y")).item(0, 0)
     stats, p = res["statistic"], res["pvalue"]
 
     df2 = df.to_pandas()
@@ -1535,10 +1569,10 @@ def test_radius_ptwise(df, dist, res):
 def test_nb_cnt(df, r, dist, res):
     test = df.select(
         pds.query_nb_cnt(
-            r,
             pl.col("x"),
             pl.col("y"),
             pl.col("z"),  # Columns used as the coordinates in n-d space
+            r=r,
             dist=dist,
         )
         .cast(pl.UInt32)
@@ -1619,9 +1653,8 @@ def test_approximate_entropy(s, m, r, scale, res):
 
 def test_approximate_entropy_edge_cases():
     df = pl.Series(name="a", values=[1]).to_frame()
-    # Error because after slicing, this becomes empty
-    with pytest.raises(Exception) as _:
-        _ = df.select(pds.query_approx_entropy("a", m=2, filtering_level=0.1, scale_by_std=False))
+    res = df.select(pds.query_approx_entropy("a", m=2, filtering_level=0.1, scale_by_std=False))
+    assert np.isnan(res.item(0, 0))
 
 
 @pytest.mark.parametrize(
@@ -1635,7 +1668,7 @@ def test_approximate_entropy_edge_cases():
     ],
 )
 def test_psi(df, n_bins, res):
-    ans = df.select(pds.query_psi("act", pl.col("ref"), n_bins=n_bins)).item(0, 0)
+    ans = df.select(pds.psi("act", pl.col("ref"), n_bins=n_bins)).item(0, 0)
     assert np.isclose(ans, res)
 
 
@@ -1663,7 +1696,7 @@ def test_psi(df, n_bins, res):
     ],
 )
 def test_psi_discrete(df, res):
-    ans = df.select(pds.query_psi_discrete("act", pl.col("ref"))).item(0, 0)
+    ans = df.select(pds.psi_discrete("act", pl.col("ref"))).item(0, 0)
     assert np.isclose(ans, res)
 
 
@@ -1704,3 +1737,122 @@ def test_query_similar_count():
         actual_cnt += int((square_l2_distance(sl, query) < 0.1))
 
     assert cnt == actual_cnt
+
+
+def test_auto_corr():
+    def autocorr(x, lag):
+        mean = np.mean(x)
+        v = np.var(x)
+        xp = x - mean
+        return np.sum(xp[lag:] * xp[:-lag]) / ((len(x) - lag) * v)
+
+    df = pds.frame(size=2_000).select(
+        pds.random(0.0, 1.0).alias("x1"),
+    )
+    x = df["x1"].to_numpy()
+
+    pds_result = df.select(
+        *(pds.query_auto_corr("x1", lag=i).alias(str(i)) for i in range(1, 10))
+    ).row(0)
+    pds_result = np.array(pds_result)
+    np_result = np.array([autocorr(x, i) for i in range(1, 10)])
+
+    assert np.all(np.abs(pds_result - np_result) < 1e-6)
+
+
+@pytest.mark.parametrize(
+    "df",
+    [
+        (
+            pl.DataFrame(
+                {
+                    "a": [1.5, 1.0, 4.0, 6.0, 5.7, 5.0, 7.8, 9.0, 7.5, 9.5, 9.0],
+                    "b": [1, 1, 2, 2, 1, 1, 3, 3, 1, 1, 1],
+                }
+            )
+        ),
+    ],
+)
+def test_isotonic(df):
+    from scipy.optimize import isotonic_regression
+
+    pds_no_weights = df.select(res=pds.isotonic_regression(pl.col("a"), weights=None))[
+        "res"
+    ].to_numpy()
+
+    scipy_no_weights = isotonic_regression(df["a"].to_numpy()).x
+
+    assert np.all(pds_no_weights == scipy_no_weights)
+
+    pds_w_weights = df.select(res=pds.isotonic_regression(pl.col("a"), weights=pl.col("b")))[
+        "res"
+    ].to_numpy()
+
+    scipy_w_weights = isotonic_regression(df["a"].to_numpy(), weights=df["b"].to_numpy()).x
+
+    assert np.all(pds_w_weights == scipy_w_weights)
+
+
+def test_next_up_down():
+    a = np.random.random(size=100)
+    df = pl.DataFrame({"a": a})
+    a_up = df.select(pds.next_up("a"))["a"].to_numpy()
+    a_down = df.select(pds.next_down("a"))["a"].to_numpy()
+
+    assert np.all(np.nextafter(a, 1.0) == a_up)
+    assert np.all(np.nextafter(a, 0.0) == a_down)
+
+def test_xlogy():
+    df = pl.DataFrame({
+        "a": [0.0, 0.0, float("nan"), 3.0], 
+        "b": [1.0, float("nan"), 1.0, 4.0], 
+    })
+    # a = 0 and b is not nan, then a * log(b) = 0
+    # otherwise, do a * log(b)
+    answer = [0.0, float("nan"), float("nan"), float(3.0 * np.log(4.0))]
+    pds_result = df.select(
+        res = pds.xlogy("a", "b")
+    )["res"].to_numpy()
+    assert np.allclose(pds_result, answer, equal_nan=True)
+
+def test_digamma():
+    import scipy
+
+    a = np.random.random(size=100)
+    df = pl.DataFrame({"a": a})
+    pds_digamma = df.select(pds.digamma(pl.col("a")))["a"].to_numpy()
+    scipy_digamma = scipy.special.psi(a)
+
+    assert np.all(np.isclose(pds_digamma, scipy_digamma, atol=1e-5))
+
+
+def test_kth_nb_dist():
+    size = 2000
+    df = pl.DataFrame(
+        {
+            "id": range(size),
+        }
+    ).with_columns(
+        pds.random().alias("var1"),
+        pds.random().alias("var2"),
+        pds.random().alias("var3"),
+    )
+    # method 1 is what we want to test
+    # method 2 is assumed to be the truth.
+    test = (
+        df.select(
+            pds.query_dist_from_kth_nb("var1", "var2", "var3", dist="l1", k=3).alias(
+                "kth_nb_dist_method_1"
+            ),
+            pds.query_knn_ptwise(
+                "var1", "var2", "var3", index="id", return_dist=True, k=3, dist="l1"
+            )
+            .struct.field("dist")
+            .list.last()
+            .alias("kth_nb_dist_method_2"),
+        )
+        .select((pl.col("kth_nb_dist_method_1") == pl.col("kth_nb_dist_method_2")).all())
+        .item(0, 0)
+    )
+
+    assert test is True
